@@ -91,7 +91,6 @@
 #include "mem/mem.h"
 #include "dprint.h"
 #include "config.h"
-#include "md5utils.h"
 #include "data_lump.h"
 #include "data_lump_rpl.h"
 #include "ip_addr.h"
@@ -103,6 +102,8 @@
 #include "parser/parse_param.h"
 #include "forward.h"
 #include "str_list.h"
+#include "pvapi.h"
+#include "xavp.h"
 #include "rand/kam_rand.h"
 
 #define append_str_trans(_dest,_src,_len,_msg) \
@@ -111,7 +112,8 @@
 extern char version[];
 extern int version_len;
 
-
+str _ksr_xavp_via_params = STR_NULL;
+str _ksr_xavp_via_fields = STR_NULL;
 
 /** per process fixup function for global_req_flags.
   * It should be called from the configuration framework.
@@ -168,12 +170,17 @@ static int check_via_address(struct ip_addr* ip, str *name,
 						(name->s[name->len-1]==']')&&
 						(strncasecmp(name->s+1, s, len)==0))
 				)
-		   )
+		   ) {
 			return 0;
-		else
-
-			if (strncmp(name->s, s, name->len)==0)
+		}
+		else {
+			if (unlikely(name->s==NULL)) {
+				LM_CRIT("invalid Via host name\n");
+				return -1;
+			}
+			if (len==name->len&&(strncmp(name->s, s, name->len)==0))
 				return 0;
+		}
 	}else{
 		LM_CRIT("could not convert ip address\n");
 		return -1;
@@ -324,7 +331,7 @@ char* received_builder(struct sip_msg *msg, unsigned int *received_len)
 	buf=pkg_malloc(sizeof(char)*MAX_RECEIVED_SIZE);
 	if (buf==0){
 		ser_error=E_OUT_OF_MEM;
-		LM_ERR("out of memory\n");
+		PKG_MEM_ERROR;
 		return 0;
 	}
 	memcpy(buf, RECEIVED, RECEIVED_LEN);
@@ -357,7 +364,7 @@ char* rport_builder(struct sip_msg *msg, unsigned int *rport_len)
 	buf=pkg_malloc(sizeof(char)*(len+1));/* space for null term */
 	if (buf==0){
 		ser_error=E_OUT_OF_MEM;
-		LM_ERR("out of memory\n");
+		PKG_MEM_ERROR;
 		return 0;
 	}
 	memcpy(buf, RPORT, RPORT_LEN);
@@ -389,7 +396,7 @@ char* id_builder(struct sip_msg* msg, unsigned int *id_len)
 	buf=pkg_malloc(sizeof(char)*(len+1));/* place for ending \0 */
 	if (buf==0){
 		ser_error=E_OUT_OF_MEM;
-		LM_ERR("out of memory\n");
+		PKG_MEM_ERROR;
 		return 0;
 	}
 	memcpy(buf, ID_PARAM, ID_PARAM_LEN);
@@ -432,7 +439,7 @@ char* clen_builder(	struct sip_msg* msg, int *clen_len, int diff,
 	buf=pkg_malloc(sizeof(char)*(len+1));
 	if (buf==0){
 		ser_error=E_OUT_OF_MEM;
-		LM_ERR("out of memory\n");
+		PKG_MEM_ERROR;
 		return 0;
 	}
 	if (body_only) {
@@ -585,7 +592,7 @@ static inline int lumps_len(struct sip_msg* msg, struct lump* lumps,
 	#define RCVCOMP_LUMP_LEN
 	#define SENDCOMP_LUMP_LEN
 #endif /*USE_COMP */
-	
+
 #define SUBST_LUMP_LEN(subst_l) \
 		switch((subst_l)->u.subst){ \
 			case SUBST_RCV_IP: \
@@ -637,9 +644,13 @@ static inline int lumps_len(struct sip_msg* msg, struct lump* lumps,
 				}; \
 				break; \
 			case SUBST_RCV_ALL: \
+			case SUBST_RCV_ALL_EX: \
 				if (msg->rcv.bind_address){ \
 					new_len+=recv_address_str->len; \
-					if (msg->rcv.bind_address->address.af!=AF_INET) \
+					if ((msg->rcv.bind_address->address.af==AF_INET6)\
+							&& (recv_address_str->s[0]!='[')\
+							&& (memchr(recv_address_str->s, ':',\
+								recv_address_str->len)!=NULL))\
 						new_len+=2; \
 					if (recv_port_no!=SIP_PORT){ \
 						/* add :port_no */ \
@@ -669,6 +680,11 @@ static inline int lumps_len(struct sip_msg* msg, struct lump* lumps,
 						LM_CRIT("unknown proto %d\n", \
 								msg->rcv.bind_address->proto); \
 					}\
+					if((subst_l)->u.subst==SUBST_RCV_ALL_EX \
+								&& msg->rcv.bind_address->sockname.len>0) { \
+						new_len+=SOCKNAME_PARAM_LEN \
+								+ msg->rcv.bind_address->sockname.len; \
+					} \
 					RCVCOMP_LUMP_LEN \
 				}else{ \
 					/* FIXME */ \
@@ -678,8 +694,8 @@ static inline int lumps_len(struct sip_msg* msg, struct lump* lumps,
 			case SUBST_SND_IP: \
 				if (send_sock){ \
 					new_len+=send_address_str->len; \
-					if (send_sock->address.af!=AF_INET && \
-							send_address_str==&(send_sock->address_str)) \
+					if (send_sock->address.af==AF_INET6 && \
+							send_address_str->s[0]!='[') \
 						new_len+=2; \
 				}else{ \
 					LM_CRIT("FIXME: null send_sock\n"); \
@@ -722,10 +738,13 @@ static inline int lumps_len(struct sip_msg* msg, struct lump* lumps,
 				}; \
 				break; \
 			case SUBST_SND_ALL: \
+			case SUBST_SND_ALL_EX: \
 				if (send_sock){ \
 					new_len+=send_address_str->len; \
-					if ((send_sock->address.af!=AF_INET) && \
-							(send_address_str==&(send_sock->address_str))) \
+					if ((send_sock->address.af==AF_INET6) && \
+							(send_address_str->s[0]!='[')\
+							&& (memchr(send_address_str->s, ':',\
+								send_address_str->len)!=NULL)) \
 						new_len+=2; \
 					if ((send_sock->port_no!=SIP_PORT) || \
 							(send_port_str!=&(send_sock->port_no_str))){ \
@@ -755,6 +774,11 @@ static inline int lumps_len(struct sip_msg* msg, struct lump* lumps,
 						default: \
 						LM_CRIT("unknown proto %d\n", send_sock->proto); \
 					}\
+					if((subst_l)->u.subst==SUBST_SND_ALL_EX \
+								&& send_sock->sockname.len>0) { \
+						new_len+=SOCKNAME_PARAM_LEN \
+								+ send_sock->sockname.len; \
+					} \
 					SENDCOMP_LUMP_LEN \
 				}else{ \
 					/* FIXME */ \
@@ -766,7 +790,7 @@ static inline int lumps_len(struct sip_msg* msg, struct lump* lumps,
 			default: \
 				LM_CRIT("unknown subst type %d\n", (subst_l)->u.subst); \
 		}
-	
+
 	if (send_info){
 		send_sock=send_info->send_sock;
 	}else{
@@ -935,7 +959,7 @@ void process_lumps( struct sip_msg* msg,
 					default:\
 						LM_CRIT("unknown comp %d\n", msg->rcv.comp); \
 				}
-	
+
 	#define SENDCOMP_PARAM_ADD \
 				/* add ;comp=xxxx */ \
 				switch(send_info->comp){ \
@@ -956,7 +980,7 @@ void process_lumps( struct sip_msg* msg,
 						break;\
 					default:\
 						LM_CRIT("unknown comp %d\n", msg->rcv.comp); \
-				} 
+				}
 #else
 	#define RCVCOMP_PARAM_ADD
 	#define SENDCOMP_PARAM_ADD
@@ -991,15 +1015,22 @@ void process_lumps( struct sip_msg* msg,
 			}; \
 			break; \
 		case SUBST_RCV_ALL: \
+		case SUBST_RCV_ALL_EX: \
 			if (msg->rcv.bind_address){  \
 				/* address */ \
-				if (msg->rcv.bind_address->address.af!=AF_INET){\
+				if ((msg->rcv.bind_address->address.af==AF_INET6)\
+						&& (recv_address_str->s[0]!='[')\
+						&& (memchr(recv_address_str->s, ':',\
+								recv_address_str->len)!=NULL)){\
 					new_buf[offset]='['; offset++; \
 				}\
 				memcpy(new_buf+offset, recv_address_str->s, \
 						recv_address_str->len); \
 				offset+=recv_address_str->len; \
-				if (msg->rcv.bind_address->address.af!=AF_INET){\
+				if ((msg->rcv.bind_address->address.af==AF_INET6)\
+						&& (recv_address_str->s[0]!='[')\
+						&& (memchr(recv_address_str->s, ':',\
+								recv_address_str->len)!=NULL)){\
 					new_buf[offset]=']'; offset++; \
 				}\
 				/* :port */ \
@@ -1048,6 +1079,15 @@ void process_lumps( struct sip_msg* msg,
 					default: \
 						LM_CRIT("unknown proto %d\n", msg->rcv.bind_address->proto); \
 				} \
+				if((subst_l)->u.subst==SUBST_RCV_ALL_EX \
+								&& msg->rcv.bind_address->sockname.len>0) { \
+					memcpy(new_buf+offset, SOCKNAME_PARAM, \
+								SOCKNAME_PARAM_LEN); \
+					offset+=SOCKNAME_PARAM_LEN; \
+					memcpy(new_buf+offset, msg->rcv.bind_address->sockname.s, \
+							msg->rcv.bind_address->sockname.len); \
+					offset+=msg->rcv.bind_address->sockname.len; \
+				} \
 				RCVCOMP_PARAM_ADD \
 			}else{  \
 				/*FIXME*/ \
@@ -1057,14 +1097,14 @@ void process_lumps( struct sip_msg* msg,
 		case SUBST_SND_IP: \
 			if (send_sock){  \
 				if ((send_sock->address.af!=AF_INET) && \
-						(send_address_str==&(send_sock->address_str))){\
+						(send_address_str->s[0]!='[')){\
 					new_buf[offset]='['; offset++; \
 				}\
 				memcpy(new_buf+offset, send_address_str->s, \
 									send_address_str->len); \
 				offset+=send_address_str->len; \
 				if ((send_sock->address.af!=AF_INET) && \
-						(send_address_str==&(send_sock->address_str))){\
+						(send_address_str->s[0]!='[')){\
 					new_buf[offset]=']'; offset++; \
 				}\
 			}else{  \
@@ -1083,17 +1123,22 @@ void process_lumps( struct sip_msg* msg,
 			}; \
 			break; \
 		case SUBST_SND_ALL: \
+		case SUBST_SND_ALL_EX: \
 			if (send_sock){  \
 				/* address */ \
-				if ((send_sock->address.af!=AF_INET) && \
-						(send_address_str==&(send_sock->address_str))){\
+				if ((send_sock->address.af==AF_INET6)\
+						&& (send_address_str->s[0]!='[')\
+						&& (memchr(send_address_str->s, ':',\
+								send_address_str->len)!=NULL)){\
 					new_buf[offset]='['; offset++; \
 				}\
 				memcpy(new_buf+offset, send_address_str->s, \
 						send_address_str->len); \
 				offset+=send_address_str->len; \
-				if ((send_sock->address.af!=AF_INET) && \
-						(send_address_str==&(send_sock->address_str))){\
+				if ((send_sock->address.af==AF_INET6)\
+						&& (send_address_str->s[0]!='[')\
+						&& (memchr(send_address_str->s, ':',\
+								send_address_str->len)!=NULL)){\
 					new_buf[offset]=']'; offset++; \
 				}\
 				/* :port */ \
@@ -1142,6 +1187,15 @@ void process_lumps( struct sip_msg* msg,
 					default: \
 						LM_CRIT("unknown proto %d\n", send_sock->proto); \
 				} \
+				if((subst_l)->u.subst==SUBST_SND_ALL_EX \
+							&& send_sock->sockname.len>0) { \
+					memcpy(new_buf+offset, SOCKNAME_PARAM, \
+							SOCKNAME_PARAM_LEN); \
+					offset+=SOCKNAME_PARAM_LEN; \
+					memcpy(new_buf+offset, send_sock->sockname.s, \
+							send_sock->sockname.len); \
+					offset+=send_sock->sockname.len; \
+				} \
 				SENDCOMP_PARAM_ADD \
 			}else{  \
 				/*FIXME*/ \
@@ -1688,16 +1742,14 @@ int get_boundary(struct sip_msg* msg, str* boundary)
 
 	params.s = memchr(msg->content_type->body.s, ';',
 		msg->content_type->body.len);
-	if (params.s == NULL)
-	{
-		LM_INFO("Content-Type hdr has no params <%.*s>\n",
+	if (params.s == NULL) {
+		LM_INFO("Content-Type hdr has no boundary params <%.*s>\n",
 				msg->content_type->body.len, msg->content_type->body.s);
-		return -1;
+		return -2;
 	}
 	params.len = msg->content_type->body.len -
 		(params.s - msg->content_type->body.s);
-	if (parse_params(&params, CLASS_ANY, &hooks, &list) < 0)
-	{
+	if (parse_params(&params, CLASS_ANY, &hooks, &list) < 0) {
 		LM_ERR("while parsing Content-Type params\n");
 		return -1;
 	}
@@ -1705,13 +1757,11 @@ int get_boundary(struct sip_msg* msg, str* boundary)
 	boundary->len = 0;
 	for (p = list; p; p = p->next) {
 		if ((p->name.len == 8)
-			&& (strncasecmp(p->name.s, "boundary", 8) == 0))
-		{
+			&& (strncasecmp(p->name.s, "boundary", 8) == 0)) {
 			boundary->s = pkg_malloc(p->body.len + 2);
-			if (boundary->s == NULL)
-			{
+			if (boundary->s == NULL) {
 				free_params(list);
-				LM_ERR("no memory for boundary string\n");
+				PKG_MEM_ERROR;
 				return -1;
 			}
 			*(boundary->s) = '-';
@@ -1742,9 +1792,17 @@ int check_boundaries(struct sip_msg *msg, struct dest_info *send_info)
 	int t, ret, lb_size;
 	char *pb;
 
-	if(!(msg->msg_flags&FL_BODY_MULTIPART)) return 0;
-	else
-	{
+	if(!(msg->msg_flags&FL_BODY_MULTIPART)) {
+		LM_DBG("no multi-part body\n");
+		return 0;
+	} else {
+		if((t = get_boundary(msg, &ob)) != 0) {
+			if(t==-2) {
+				LM_INFO("no boundary - maybe just turning into multipart body\n");
+				return -2;
+			}
+			return -1;
+		}
 		buf.s = build_body(msg, (unsigned int *)&buf.len, &ret, send_info);
 		if(ret) {
 			LM_ERR("Can't get body\n");
@@ -1752,10 +1810,6 @@ int check_boundaries(struct sip_msg *msg, struct dest_info *send_info)
 		}
 		tmp.s = buf.s;
 		t = tmp.len = buf.len;
-		if(get_boundary(msg, &ob)!=0) {
-			if(tmp.s) pkg_free(tmp.s);
-			return -1;
-		}
 		if(str_append(&ob, &bsuffix, &b)!=0) {
 			LM_ERR("Can't append suffix to boundary\n");
 			goto error;
@@ -1815,10 +1869,10 @@ int check_boundaries(struct sip_msg *msg, struct dest_info *send_info)
 			tmp.len = get_line(lb_t->s);
 			if(tmp.len!=b.len || strncmp(b.s, tmp.s, b.len)!=0)
 			{
-				LM_DBG("malformed bondary in the middle\n");
+				LM_DBG("malformed boundary in the middle\n");
 				memcpy(pb, b.s, b.len); body.len = body.len + b.len;
 				pb = pb + b.len;
-				t = lb_t->s.s - (lb_t->s.s + tmp.len);
+				t = lb_t->next->s.s - (lb_t->s.s + tmp.len);
 				memcpy(pb, lb_t->s.s+tmp.len, t); pb = pb + t;
 				/*LM_DBG("new chunk[%d][%.*s]\n", t, t, pb-t);*/
 			}
@@ -1939,6 +1993,7 @@ char * build_req_buf_from_sip_req( struct sip_msg* msg,
 	unsigned int flags;
 	unsigned int udp_mtu;
 	struct dest_info di;
+	int ret;
 
 	via_insert_param=0;
 	uri_len=0;
@@ -1956,9 +2011,10 @@ char * build_req_buf_from_sip_req( struct sip_msg* msg,
 	path_buf.len=0;
 
 	flags=msg->msg_flags|global_req_flags;
-	if(check_boundaries(msg, send_info)<0){
-		LM_WARN("check_boundaries error\n");
+	if((ret = check_boundaries(msg, send_info)) < 0){
+		LM_INFO("check boundaries negative (%d)\n", ret);
 	}
+
 	/* Calculate message body difference and adjust Content-Length */
 	body_delta = lumps_len(msg, msg->body_lumps, send_info);
 	if (adjust_clen(msg, body_delta, send_info->proto) < 0) {
@@ -1987,7 +2043,7 @@ after_local_via:
 	if ( received_test(msg) ) {
 		if ((received_buf=received_builder(msg,&received_len))==0){
 			LM_ERR("received_builder failed\n");
-			goto error01;  /* free also line_buf */
+			goto error00;  /* free also line_buf */
 		}
 	}
 
@@ -2000,7 +2056,7 @@ after_local_via:
 			(msg->via1->rport /*&& msg->via1->rport->value.s==0*/)){
 		if ((rport_buf=rport_builder(msg, &rport_len))==0){
 			LM_ERR("rport_builder failed\n");
-			goto error01; /* free everything */
+			goto error00; /* free everything */
 		}
 	}
 
@@ -2022,7 +2078,7 @@ after_local_via:
 	}
 	/* if received needs to be added, add anchor after host and add it, or
 	 * overwrite the previous one if already present */
-	if (received_len){
+	if (received_buf){
 		if (msg->via1->received){ /* received already present => overwrite it*/
 			via_insert_param=del_lump(msg,
 								msg->via1->received->start-buf-1, /*;*/
@@ -2031,12 +2087,15 @@ after_local_via:
 			via_insert_param=anchor_lump(msg, msg->via1->hdr.s-buf+size, 0,
 											HDR_VIA_T);
 		}
-		if (via_insert_param==0) goto error02; /* free received_buf */
+		if (via_insert_param==0) goto error00; /* free received_buf */
 		if (insert_new_lump_after(via_insert_param, received_buf, received_len,
-					HDR_VIA_T) ==0 ) goto error02; /* free received_buf */
+					HDR_VIA_T) ==0 ) {
+			goto error00; /* free received_buf */
+		}
+		received_buf = NULL;
 	}
 	/* if rport needs to be updated, delete it if present and add it's value */
-	if (rport_len){
+	if (rport_buf){
 		if (msg->via1->rport){ /* rport already present */
 			via_insert_param=del_lump(msg,
 								msg->via1->rport->start-buf-1, /*';'*/
@@ -2046,11 +2105,12 @@ after_local_via:
 			via_insert_param=anchor_lump(msg, msg->via1->hdr.s-buf+size, 0,
 											HDR_VIA_T);
 		}
-		if (via_insert_param==0) goto error03; /* free rport_buf */
+		if (via_insert_param==0) goto error00; /* free rport_buf */
 		if (insert_new_lump_after(via_insert_param, rport_buf, rport_len,
-									HDR_VIA_T) ==0 )
-			goto error03; /* free rport_buf */
-
+									HDR_VIA_T) ==0 ) {
+			goto error00; /* free rport_buf */
+		}
+		rport_buf = NULL;
 	}
 
 after_update_via1:
@@ -2060,7 +2120,7 @@ after_update_via1:
 		path_buf.len=ROUTE_PREFIX_LEN+msg->path_vec.len+CRLF_LEN;
 		path_buf.s=pkg_malloc(path_buf.len+1);
 		if (unlikely(path_buf.s==0)){
-			LM_ERR("out of memory\n");
+		        PKG_MEM_ERROR;
 			ser_error=E_OUT_OF_MEM;
 			goto error00;
 		}
@@ -2073,25 +2133,26 @@ after_update_via1:
 		   (if present & parsed), after the local via or after in front of
 		    the first via if we don't add a local via*/
 		if (msg->route){
-			path_anchor=anchor_lump(msg, msg->route->name.s-buf, 0, 
+			path_anchor=anchor_lump(msg, msg->route->name.s-buf, 0,
 									HDR_ROUTE_T);
 		}else if (likely(via_anchor)){
 			path_anchor=via_anchor;
 		}else if (likely(msg->via1)){
-			path_anchor=anchor_lump(msg, msg->via1->hdr.s-buf, 0, 
+			path_anchor=anchor_lump(msg, msg->via1->hdr.s-buf, 0,
 									HDR_ROUTE_T);
 		}else{
 			/* if no via1 (theoretically possible for non-sip messages,
 			   e.g. http xmlrpc) */
-			path_anchor=anchor_lump(msg, msg->headers->name.s-buf, 0, 
+			path_anchor=anchor_lump(msg, msg->headers->name.s-buf, 0,
 									HDR_ROUTE_T);
 		}
 		if (unlikely(path_anchor==0))
-			goto error05;
+			goto error00;
 		if (unlikely((path_lump=insert_new_lump_after(path_anchor, path_buf.s,
-														path_buf.len,
-														HDR_ROUTE_T))==0))
-			goto error05;
+									path_buf.len, HDR_ROUTE_T))==0)) {
+			goto error00;
+		}
+		path_buf.s = NULL;
 	}
 	/* compute new msg len and fix overlapping zones*/
 	new_len=len+body_delta+lumps_len(msg, msg->add_rm, send_info)+via_len;
@@ -2100,7 +2161,7 @@ after_update_via1:
 #endif
 	udp_mtu=cfg_get(core, core_cfg, udp_mtu);
 	di.proto=PROTO_NONE;
-	if (unlikely((send_info->proto==PROTO_UDP) && udp_mtu && 
+	if (unlikely((send_info->proto==PROTO_UDP) && udp_mtu &&
 					(flags & FL_MTU_FB_MASK) && (new_len>udp_mtu)
 					&& (!(mode&BUILD_NO_LOCAL_VIA)))){
 
@@ -2127,7 +2188,7 @@ after_update_via1:
 			di.proto=PROTO_SCTP;
 		 }
 #endif /* USE_SCTP */
-		
+
 		if (di.proto!=PROTO_NONE){
 			new_len-=via_len;
 			if(likely(line_buf)) pkg_free(line_buf);
@@ -2144,8 +2205,10 @@ after_update_via1:
 	/* add first via, as an anchor for second via*/
 	if(likely(line_buf)) {
 		if ((via_lump=insert_new_lump_before(via_anchor, line_buf, via_len,
-											HDR_VIA_T))==0)
-			goto error04;
+											HDR_VIA_T))==0) {
+			goto error00;
+		}
+		line_buf = 0;
 	}
 	if (msg->new_uri.s){
 		uri_len=msg->new_uri.len;
@@ -2157,7 +2220,11 @@ after_update_via1:
 		new_buf=(char*)pkg_malloc(new_len+1);
 	if (new_buf==0){
 		ser_error=E_OUT_OF_MEM;
-		LM_ERR("out of memory\n");
+		if(unlikely(mode&BUILD_IN_SHM)) {
+			SHM_MEM_ERROR;
+		} else {
+			PKG_MEM_ERROR;
+		}
 		goto error00;
 	}
 
@@ -2182,7 +2249,7 @@ after_update_via1:
 	new_buf[new_len]=0;
 
 	/* update the send_info if udp_mtu affected */
-	if (di.proto!=PROTO_NONE) { 
+	if (di.proto!=PROTO_NONE) {
 		send_info->proto=di.proto;
 		send_info->send_sock=di.send_sock;
 	}
@@ -2197,16 +2264,12 @@ after_update_via1:
 	*returned_len=new_len;
 	return new_buf;
 
-error01:
-error02:
-	if (received_buf) pkg_free(received_buf);
-error03:
-	if (rport_buf) pkg_free(rport_buf);
-error04:
-	if (line_buf) pkg_free(line_buf);
-error05:
-	if (path_buf.s) pkg_free(path_buf.s);
 error00:
+	if (received_buf) pkg_free(received_buf);
+	if (rport_buf) pkg_free(rport_buf);
+	if (path_buf.s) pkg_free(path_buf.s);
+	if (line_buf) pkg_free(line_buf);
+
 	*returned_len=0;
 	return 0;
 }
@@ -2263,7 +2326,7 @@ char * generate_res_buf_from_sip_res( struct sip_msg* msg,
 	new_buf=(char*)pkg_malloc(new_len+1); /* +1 is for debugging
 											 (\0 to print it )*/
 	if (new_buf==0){
-		LM_ERR("out of mem\n");
+	        PKG_MEM_ERROR;
 		goto error;
 	}
 	new_buf[new_len]=0; /* debug: print the message */
@@ -2364,7 +2427,7 @@ char * build_res_buf_from_sip_req( unsigned int code, str *text ,str *new_tag,
 			case HDR_TO_T:
 				if (new_tag && new_tag->len) {
 					to_tag=get_to(msg)->tag_value;
-					if ( to_tag.len || to_tag.s )
+					if ( to_tag.len && to_tag.s )
 						len+=new_tag->len-to_tag.len;
 					else
 						len+=new_tag->len+TOTAG_TOKEN_LEN/*";tag="*/;
@@ -2420,7 +2483,7 @@ char * build_res_buf_from_sip_req( unsigned int code, str *text ,str *new_tag,
 	buf = (char*) pkg_malloc( len+1 );
 	if (!buf)
 	{
-		LM_ERR("out of memory; needs %d\n",len);
+	        PKG_MEM_ERROR;
 		goto error01;
 	}
 
@@ -2492,7 +2555,7 @@ char * build_res_buf_from_sip_req( unsigned int code, str *text ,str *new_tag,
 				break;
 			case HDR_TO_T:
 				if (new_tag && new_tag->len){
-					if (to_tag.s ) { /* replacement */
+					if (to_tag.len && to_tag.s) { /* replacement */
 						/* before to-tag */
 						append_str( p, hdr->name.s, to_tag.s-hdr->name.s);
 						/* to tag replacement */
@@ -2648,10 +2711,10 @@ int branch_builder( unsigned int hash_index,
 
 
 
-/* uses only the send_info->send_socket, send_info->proto and 
+/* uses only the send_info->send_socket, send_info->proto and
  * send_info->comp (so that a send_info used for sending can be passed
  * to this function w/o changes and the correct via will be built) */
-char* via_builder( unsigned int *len,
+char* via_builder(unsigned int *len, sip_msg_t *msg,
 	struct dest_info* send_info /* where to send the reply */,
 	str* branch, str* extra_params, struct hostport* hp)
 {
@@ -2659,8 +2722,8 @@ char* via_builder( unsigned int *len,
 	char               *line_buf;
 	int max_len;
 	int via_prefix_len;
-	str* address_str; /* address displayed in via */
-	str* port_str; /* port no displayed in via */
+	str* address_str = NULL; /* address displayed in via */
+	str* port_str = NULL; /* port no displayed in via */
 	struct socket_info* send_sock;
 	int comp_len, comp_name_len;
 #ifdef USE_COMP
@@ -2671,22 +2734,46 @@ char* via_builder( unsigned int *len,
 	union sockaddr_union *from = NULL;
 	union sockaddr_union local_addr;
 	struct tcp_connection *con = NULL;
+	sr_xavp_t *rxavp = NULL;
+	str xname;
 
 	send_sock=send_info->send_sock;
 	/* use pre-set address in via, the outbound socket alias or address one */
-	if (hp && hp->host->len)
-		address_str=hp->host;
-	else if(send_sock->useinfo.name.len>0)
-		address_str=&(send_sock->useinfo.name);
-	else
-		address_str=&(send_sock->address_str);
-	if (hp && hp->port->len)
-		port_str=hp->port;
-	else if(send_sock->useinfo.port_no>0)
-		port_str=&(send_sock->useinfo.port_no_str);
-	else
-		port_str=&(send_sock->port_no_str);
-	
+	if(msg && (msg->msg_flags&FL_USE_XAVP_VIA_FIELDS)
+			&& _ksr_xavp_via_fields.len>0) {
+		xname.s = "address";
+		xname.len = 7;
+		rxavp = xavp_get_child_with_sval(&_ksr_xavp_via_fields, &xname);
+		if(rxavp!=NULL) {
+			address_str = &rxavp->val.v.s;
+		}
+	}
+	if(address_str==NULL) {
+		if (hp && hp->host->len)
+			address_str=hp->host;
+		else if(send_sock->useinfo.name.len>0)
+			address_str=&(send_sock->useinfo.name);
+		else
+			address_str=&(send_sock->address_str);
+	}
+	if(msg && (msg->msg_flags&FL_USE_XAVP_VIA_FIELDS)
+			&& _ksr_xavp_via_fields.len>0) {
+		xname.s = "port";
+		xname.len = 4;
+		rxavp = xavp_get_child_with_sval(&_ksr_xavp_via_fields, &xname);
+		if(rxavp!=NULL) {
+			port_str = &rxavp->val.v.s;
+		}
+	}
+	if(port_str==NULL) {
+		if (hp && hp->port->len)
+			port_str=hp->port;
+		else if(send_sock->useinfo.port_no>0)
+			port_str=&(send_sock->useinfo.port_no_str);
+		else
+			port_str=&(send_sock->port_no_str);
+	}
+
 	comp_len=comp_name_len=0;
 #ifdef USE_COMP
 	comp_name=0;
@@ -2720,7 +2807,7 @@ char* via_builder( unsigned int *len,
 	line_buf=pkg_malloc( max_len );
 	if (line_buf==0){
 		ser_error=E_OUT_OF_MEM;
-		LM_ERR("out of memory\n");
+		PKG_MEM_ERROR;
 		return 0;
 	}
 
@@ -2782,14 +2869,18 @@ char* via_builder( unsigned int *len,
 		pkg_free(line_buf);
 		return 0;
 	}
-	/* add [] only if ipv6 and outbound socket address is used;
+	/* add [] only if ipv6 address is used;
 	 * if using pre-set no check is made */
-	if ((send_sock->address.af==AF_INET6) &&
-		(address_str==&(send_sock->address_str))) {
-		line_buf[via_prefix_len]='[';
-		line_buf[via_prefix_len+1+address_str->len]=']';
-		extra_len=1;
-		via_len+=2; /* [ ]*/
+	if (send_sock->address.af==AF_INET6) {
+		/* lightweight safety checks if brackets set
+		 * or non-ipv6 (e.g., advertised hostname) */
+		if(address_str->s[0] != '['
+				&& memchr(address_str->s, ':', address_str->len)!=NULL) {
+			line_buf[via_prefix_len]='[';
+			line_buf[via_prefix_len+1+address_str->len]=']';
+			extra_len=1;
+			via_len+=2; /* [ ]*/
+		}
 	}
 	memcpy(line_buf+via_prefix_len+extra_len, address_str->s,
 				address_str->len);
@@ -2830,9 +2921,9 @@ char* via_builder( unsigned int *len,
 	return line_buf;
 }
 
-/* creates a via header honoring the protocol of the incomming socket
+/* creates a via header honoring the protocol of the incoming socket
  * msg is an optional parameter */
-char* create_via_hf( unsigned int *len,
+char* create_via_hf(unsigned int *len,
 	struct sip_msg *msg,
 	struct dest_info* send_info /* where to send the reply */,
 	str* branch)
@@ -2840,6 +2931,9 @@ char* create_via_hf( unsigned int *len,
 	char* via;
 	str extra_params;
 	struct hostport hp;
+	char sbuf[24];
+	int slen;
+	str xparams;
 #if defined USE_TCP || defined USE_SCTP
 	char* id_buf;
 	unsigned int id_len;
@@ -2885,7 +2979,7 @@ char* create_via_hf( unsigned int *len,
 		/* params so far + ';rport' + '\0' */
 		via = (char*)pkg_malloc(extra_params.len+RPORT_LEN);
 		if(via==0) {
-			LM_ERR("building local rport via param failed\n");
+		        PKG_MEM_ERROR;
 			if (extra_params.s) pkg_free(extra_params.s);
 			return 0;
 		}
@@ -2899,8 +2993,58 @@ char* create_via_hf( unsigned int *len,
 		extra_params.s[extra_params.len]='\0';
 	}
 
+	/* test and add srvid parameter to local via  */
+	if(msg && (msg->msg_flags&FL_ADD_SRVID) && server_id!=0) {
+		slen = snprintf(sbuf, 24, ";srvid=%u", (unsigned int)server_id);
+		if(slen<=0 || slen>=24) {
+			LM_WARN("failed to build srvid parameter");
+		} else {
+			via = (char*)pkg_malloc(extra_params.len+slen+1);
+			if(via==0) {
+			        PKG_MEM_ERROR;
+				if (extra_params.s) pkg_free(extra_params.s);
+				return 0;
+			}
+			if(extra_params.len != 0) {
+				memcpy(via, extra_params.s, extra_params.len);
+				pkg_free(extra_params.s);
+			}
+			memcpy(via + extra_params.len, sbuf, slen);
+			extra_params.s = via;
+			extra_params.len += slen;
+			extra_params.s[extra_params.len] = '\0';
+		}
+	}
+
+	/* test and add xavp params */
+	if(msg && (msg->msg_flags&FL_ADD_XAVP_VIA_PARAMS)
+			&& _ksr_xavp_via_params.len>0) {
+		xparams.s = pv_get_buffer();
+		xparams.len = xavp_serialize_fields(&_ksr_xavp_via_params,
+							xparams.s, pv_get_buffer_size());
+		if(xparams.len>0) {
+			via = (char*)pkg_malloc(extra_params.len+xparams.len+2);
+			if(via==0) {
+			        PKG_MEM_ERROR;
+				if (extra_params.s) pkg_free(extra_params.s);
+				return 0;
+			}
+			if(extra_params.len != 0) {
+				memcpy(via, extra_params.s, extra_params.len);
+				pkg_free(extra_params.s);
+			}
+			/* add ';' between via parameters */
+			via[extra_params.len] = ';';
+			/* skip last ';' from xavp serialized output */
+			memcpy(via + extra_params.len + 1, xparams.s, xparams.len - 1);
+			extra_params.s = via;
+			extra_params.len += xparams.len;
+			extra_params.s[extra_params.len] = '\0';
+		}
+	}
+
 	set_hostport(&hp, msg);
-	via = via_builder( len, send_info, branch,
+	via = via_builder(len, msg, send_info, branch,
 							extra_params.len?&extra_params:0, &hp);
 
 	/* we do not need extra_params any more, already in the new via header */
@@ -2941,7 +3085,7 @@ char * build_only_headers( struct sip_msg* msg, int skip_first_line,
 
 	new_buf = (char *)pkg_malloc(new_len+1);
 	if (!new_buf) {
-		LM_ERR("Not enough memory\n");
+	        PKG_MEM_ERROR;
 		*error = -1;
 		return 0;
 	}
@@ -2991,7 +3135,7 @@ char * build_body( struct sip_msg* msg,
 
 	new_buf = (char *)pkg_malloc(new_len+1);
 	if (!new_buf) {
-		LM_ERR("Not enough memory\n");
+	        PKG_MEM_ERROR;
 		*error = -1;
 		return 0;
 	}
@@ -3051,7 +3195,7 @@ char * build_all( struct sip_msg* msg, int touch_clen,
 
 	new_buf = (char *)pkg_malloc(new_len+1);
 	if (!new_buf) {
-		LM_ERR("Not enough memory\n");
+	        PKG_MEM_ERROR;
 		*error = -1;
 		return 0;
 	}
@@ -3084,6 +3228,7 @@ int build_sip_msg_from_buf(struct sip_msg *msg, char *buf, int len,
 
 	memset(msg, 0, sizeof(sip_msg_t));
 	msg->id = id;
+	msg->pid = my_pid();
 	msg->buf = buf;
 	msg->len = len;
 	if (parse_msg(buf, len, msg)!=0) {
@@ -3095,3 +3240,107 @@ int build_sip_msg_from_buf(struct sip_msg *msg, char *buf, int len,
 	return 0;
 }
 
+/**
+ *
+ */
+int sip_msg_update_buffer(sip_msg_t *msg, str *obuf)
+{
+	sip_msg_t tmp;
+
+	if(obuf==NULL || obuf->s==NULL || obuf->len<=0) {
+		LM_ERR("invalid buffer parameter\n");
+		return -1;
+	}
+
+	if(obuf->len >= BUF_SIZE) {
+		LM_ERR("new buffer is too large (%d)\n", obuf->len);
+		return -1;
+	}
+	/* temporary copy */
+	memcpy(&tmp, msg, sizeof(sip_msg_t));
+
+	/* reset dst uri and path vector to avoid freeing - restored later */
+	if(msg->dst_uri.s != NULL) {
+		msg->dst_uri.s = NULL;
+		msg->dst_uri.len = 0;
+	}
+	if(msg->path_vec.s != NULL) {
+		msg->path_vec.s = NULL;
+		msg->path_vec.len = 0;
+	}
+
+	/* free old msg structure */
+	free_sip_msg(msg);
+	memset(msg, 0, sizeof(sip_msg_t));
+
+	/* restore msg fields */
+	msg->buf = tmp.buf;
+	msg->id = tmp.id;
+	msg->rcv = tmp.rcv;
+	msg->set_global_address = tmp.set_global_address;
+	msg->set_global_port = tmp.set_global_port;
+	msg->flags = tmp.flags;
+	msg->msg_flags = tmp.msg_flags;
+	memcpy(msg->xflags, tmp.xflags, KSR_XFLAGS_SIZE * sizeof(flag_t));
+	msg->hash_index = tmp.hash_index;
+	msg->force_send_socket = tmp.force_send_socket;
+	msg->fwd_send_flags = tmp.fwd_send_flags;
+	msg->rpl_send_flags = tmp.rpl_send_flags;
+	msg->dst_uri = tmp.dst_uri;
+	msg->path_vec = tmp.path_vec;
+
+	memcpy(msg->buf, obuf->s, obuf->len);
+	msg->len = obuf->len;
+	msg->buf[msg->len] = '\0';
+
+	/* reparse the message */
+	LM_DBG("SIP message content updated - reparsing\n");
+	if(parse_msg(msg->buf, msg->len, msg) != 0) {
+		LM_ERR("parsing new sip message failed [[%.*s]]\n", msg->len, msg->buf);
+		/* exit config execution - sip_msg_t structure is no longer
+		 * valid/safe for config */
+		return 0;
+	}
+
+	return 1;
+}
+
+/**
+ *
+ */
+int sip_msg_apply_changes(sip_msg_t *msg)
+{
+	int ret;
+	dest_info_t dst;
+	str obuf;
+
+	if(msg->first_line.type != SIP_REPLY && get_route_type() != REQUEST_ROUTE) {
+		LM_ERR("invalid usage - not in request route or a reply\n");
+		return -1;
+	}
+
+	init_dest_info(&dst);
+	dst.proto = PROTO_UDP;
+	if(msg->first_line.type == SIP_REPLY) {
+		obuf.s = generate_res_buf_from_sip_res(
+				msg, (unsigned int *)&obuf.len, BUILD_NO_VIA1_UPDATE);
+	} else {
+		if(msg->msg_flags & FL_RR_ADDED) {
+			LM_ERR("cannot apply msg changes after adding record-route"
+				   " header - it breaks conditional 2nd header\n");
+			return -1;
+		}
+		obuf.s = build_req_buf_from_sip_req(msg, (unsigned int *)&obuf.len,
+				&dst,
+				BUILD_NO_PATH | BUILD_NO_LOCAL_VIA | BUILD_NO_VIA1_UPDATE);
+	}
+	if(obuf.s == NULL) {
+		LM_ERR("couldn't update msg buffer content\n");
+		return -1;
+	}
+	ret = sip_msg_update_buffer(msg, &obuf);
+	/* free new buffer - copied in the static buffer from old sip_msg_t */
+	pkg_free(obuf.s);
+
+	return ret;
+}
