@@ -39,7 +39,6 @@
 #include <unistd.h>
 #include <stdlib.h> /* for abort() */
 
-
 #include "dprint.h"
 #include "tcp_conn.h"
 #include "tcp_read.h"
@@ -70,6 +69,7 @@
 #include "forward.h"
 #include "events.h"
 #include "stun.h"
+#include "nonsip_hooks.h"
 
 #ifdef READ_HTTP11
 #define HTTP11CONTINUE	"HTTP/1.1 100 Continue\r\nContent-Length: 0\r\n\r\n"
@@ -91,6 +91,8 @@ static ticks_t tcp_reader_prev_ticks;
 
 int is_msg_complete(struct tcp_req* r);
 
+int ksr_tcp_accept_hep3=0;
+int ksr_tcp_accept_haproxy=0;
 /**
  * control cloning of TCP receive buffer
  * - needed for operations working directly inside the buffer
@@ -190,16 +192,18 @@ static int tcp_emit_closed_event(struct tcp_connection *con, enum tcp_closed_rea
 {
 	int ret;
 	tcp_closed_event_info_t tev;
+	sr_event_param_t evp = {0};
 
 	ret = 0;
-	LM_DBG("TCP closed event creation triggered\n");
+	LM_DBG("TCP closed event creation triggered (reason: %d)\n", reason);
 	if(likely(sr_event_enabled(SREV_TCP_CLOSED))) {
 		memset(&tev, 0, sizeof(tcp_closed_event_info_t));
 		tev.reason = reason;
 		tev.con = con;
-		ret = sr_event_exec(SREV_TCP_CLOSED, (void*)(&tev));
+		evp.data = (void*)(&tev);
+		ret = sr_event_exec(SREV_TCP_CLOSED, &evp);
 	} else {
-		LM_DBG("no callback registering for handling TCP closed event - dropping!\n");
+		LM_DBG("no callback registering for handling TCP closed event\n");
 	}
 	return ret;
 }
@@ -237,16 +241,17 @@ int tcp_read_data(int fd, struct tcp_connection *c,
 					char* buf, int b_size, int* flags)
 {
 	int bytes_read;
-	
+
 again:
 	bytes_read=read(fd, buf, b_size);
-	
+
 	if (likely(bytes_read!=b_size)){
 		if(unlikely(bytes_read==-1)){
 			if (errno == EWOULDBLOCK || errno == EAGAIN){
 				bytes_read=0; /* nothing has been read */
-			}else if (errno == EINTR) goto again;
-			else{
+			}else if (errno == EINTR){
+				goto again;
+			}else{
 				if (unlikely(c->state==S_CONN_CONNECT)){
 					switch(errno){
 						case ECONNRESET:
@@ -276,6 +281,12 @@ again:
 						switch(errno){
 							case ECONNRESET:
 								TCP_STATS_CON_RESET();
+#ifdef USE_DST_BLACKLIST
+								dst_blacklist_su(BLST_ERR_SEND, c->rcv.proto,
+													&c->rcv.src_su,
+													&c->send_flags, 0);
+#endif /* USE_DST_BLACKLIST */
+								break;
 							case ETIMEDOUT:
 #ifdef USE_DST_BLACKLIST
 								dst_blacklist_su(BLST_ERR_SEND, c->rcv.proto,
@@ -289,7 +300,8 @@ again:
 						"error reading: %s (%d) ([%s]:%u ->",
 						strerror(errno), errno,
 						ip_addr2a(&c->rcv.src_ip), c->rcv.src_port);
-				LOG(cfg_get(core, core_cfg, corelog),"-> [%s]:%u)\n", ip_addr2a(&c->rcv.dst_ip), c->rcv.dst_port);
+				LOG(cfg_get(core, core_cfg, corelog),"-> [%s]:%u)\n",
+						ip_addr2a(&c->rcv.dst_ip), c->rcv.dst_port);
 				if (errno == ETIMEDOUT) {
 					tcp_emit_closed_event(c, TCP_CLOSED_TIMEOUT);
 				} else if (errno == ECONNRESET) {
@@ -297,12 +309,13 @@ again:
 				}
 				return -1;
 			}
-		}else if (unlikely((bytes_read==0) || 
+		}else if (unlikely((bytes_read==0) ||
 					(*flags & RD_CONN_FORCE_EOF))){
 			c->state=S_CONN_EOF;
 			*flags|=RD_CONN_EOF;
 			tcp_emit_closed_event(c, TCP_CLOSED_EOF);
-			LM_DBG("EOF on %p, FD %d ([%s]:%u ->", c, fd, ip_addr2a(&c->rcv.src_ip), c->rcv.src_port);
+			LM_DBG("EOF on %p, FD %d ([%s]:%u ->", c, fd,
+					ip_addr2a(&c->rcv.src_ip), c->rcv.src_port);
 			LM_DBG("-> [%s]:%u)\n", ip_addr2a(&c->rcv.dst_ip), c->rcv.dst_port);
 		}else{
 			if (unlikely(c->state==S_CONN_CONNECT || c->state==S_CONN_ACCEPT)){
@@ -326,9 +339,9 @@ again:
 /* reads next available bytes
  *   c- tcp connection used for reading, tcp_read changes also c->state on
  *      EOF and c->req.error on read error
- *   * flags - value/result - used to signal a seen or "forced" EOF on the 
- *     connection (when it is known that no more data will come after the 
- *     current socket buffer is emptied )=> return/signal EOF on the first 
+ *   * flags - value/result - used to signal a seen or "forced" EOF on the
+ *     connection (when it is known that no more data will come after the
+ *     current socket buffer is emptied )=> return/signal EOF on the first
  *     short read (=> don't use it on POLLPRI, as OOB data will cause short
  *      reads even if there are still remaining bytes in the socket buffer)
  * return number of bytes read, 0 on EOF or -1 on error,
@@ -344,9 +357,9 @@ int tcp_read(struct tcp_connection *c, int* flags)
 
 	r=&c->req;
 	fd=c->fd;
-	bytes_free=r->b_size- (int)(r->pos - r->buf);
-	
-	if (unlikely(bytes_free==0)){
+	bytes_free=r->b_size - (unsigned int)(r->pos - r->buf);
+
+	if (unlikely(bytes_free<=0)){
 		LM_ERR("buffer overrun, dropping ([%s]:%u -> [%s]:%u)\n",
 				ip_addr2a(&c->rcv.src_ip), c->rcv.src_port,
 				ip_addr2a(&c->rcv.dst_ip), c->rcv.dst_port);
@@ -422,7 +435,7 @@ int tcp_read_headers(struct tcp_connection *c, int* read_flags)
 							r->state=(newstate); break; \
 						crlf_default_skip_case; \
 					}
-	
+
 	#define change_state_case(state0, upper, lower, newstate)\
 					case state0: \
 							  change_state(upper, lower, newstate); \
@@ -431,6 +444,22 @@ int tcp_read_headers(struct tcp_connection *c, int* read_flags)
 
 
 	r=&c->req;
+	if(r->parsed<r->buf || r->parsed>r->buf+r->b_size) {
+		if(r->parsed<r->buf && (unsigned char)r->state==H_SKIP_EMPTY) {
+			/* give it a chance to parse from beginning */
+			LM_WARN("resetting parsed pointer (buf:%p parsed:%p bsize:%u)\n",
+				r->buf, r->parsed, r->b_size);
+			r->parsed = r->buf;
+		} else {
+			LM_ERR("out of bounds parsed pointer (buf:%p parsed:%p bsize:%u)\n",
+				r->buf, r->parsed, r->b_size);
+			r->parsed = r->buf;
+			r->content_len=0;
+			r->error=TCP_REQ_BAD_LEN;
+			r->state=H_SKIP; /* skip state now */
+			return -1;
+		}
+	}
 	/* if we still have some unparsed part, parse it first, don't do the read*/
 	if (unlikely(r->parsed<r->pos)){
 		bytes=0;
@@ -457,7 +486,7 @@ int tcp_read_headers(struct tcp_connection *c, int* read_flags)
 					goto skip;
 				}
 				break;
-				
+
 			case H_SKIP:
 				/* find lf, we are in this state if we are not interested
 				 * in anything till end of line*/
@@ -650,10 +679,10 @@ int tcp_read_headers(struct tcp_connection *c, int* read_flags)
 						r->start=p;
 						break;
 					default:
-						/* stun test */						
+						/* stun test */
 						if (unlikely(sr_event_enabled(SREV_STUN_IN)) && (unsigned char)*p == 0x00) {
 							r->state=H_STUN_MSG;
-						/* body will used as pointer to the last used byte */
+							/* body is used as pointer to the last used byte */
 							r->body=p;
 							r->content_len = 0;
 							LM_DBG("stun msg detected\n");
@@ -1052,6 +1081,7 @@ int msrp_process_msg(char* tcpbuf, unsigned int len,
 {
 	int ret;
 	tcp_event_info_t tev;
+	sr_event_param_t evp = {0};
 
 	ret = 0;
 	LM_DBG("MSRP Message: [[>>>\n%.*s<<<]]\n", len, tcpbuf);
@@ -1062,7 +1092,8 @@ int msrp_process_msg(char* tcpbuf, unsigned int len,
 		tev.len = len;
 		tev.rcv = rcv_info;
 		tev.con = con;
-		ret = sr_event_exec(SREV_TCP_MSRP_FRAME, (void*)(&tev));
+		evp.data = (void*)(&tev);
+		ret = sr_event_exec(SREV_TCP_MSRP_FRAME, &evp);
 	} else {
 		LM_DBG("no callback registering for handling MSRP - dropping!\n");
 	}
@@ -1073,8 +1104,8 @@ int msrp_process_msg(char* tcpbuf, unsigned int len,
 #ifdef READ_WS
 static int tcp_read_ws(struct tcp_connection *c, int* read_flags)
 {
-	int bytes, size, pos, mask_present;
-	unsigned int len;
+	int bytes;
+	uint32_t size, pos, mask_present, len;
 	char *p;
 	struct tcp_req *r;
 
@@ -1086,10 +1117,28 @@ static int tcp_read_ws(struct tcp_connection *c, int* read_flags)
 #endif
 		bytes = tcp_read(c, read_flags);
 
-	if (bytes <= 0)
-	{
-		if (likely(r->parsed >= r->pos))
-			return 0;
+	if (bytes < 0) {
+		/* read error */
+		return bytes;
+	}
+	if (r->parsed == r->pos) {
+		/* nothing else to parse */
+		return bytes;
+	}
+	if (r->parsed > r->pos) {
+		LM_ERR("req buf pos (%p) before parsed (%p) [%d]\n", r->pos, r->parsed,
+				bytes);
+		return -1;
+	}
+	if(r->pos > r->buf + r->b_size) {
+		LM_ERR("req pos (%p) over buf (%p / %u) - parsed (%p) [%d]\n", r->pos,
+				r->buf, r->b_size, r->parsed, bytes);
+		return -1;
+	}
+	if(r->buf > r->parsed) {
+		LM_ERR("req parsed (%p) before buf (%p / %u) - pos (%p) [%d]\n",
+				r->parsed, r->buf, r->b_size, r->pos, bytes);
+		return -1;
 	}
 
 	size = r->pos - r->parsed;
@@ -1126,27 +1175,32 @@ static int tcp_read_ws(struct tcp_connection *c, int* read_flags)
 		goto skip;
 	pos++;
 	mask_present = p[pos] & 0x80;
-	len = (p[pos++] & 0xff) & ~0x80;
+	len = (p[pos] & 0xff) & ~0x80;
+	pos++;
 
 	/* Work out real length */
-	if (len == 126)
-	{
+	if (len == 126) {
+		/* 2 bytes store the payload size */
 		if (size < pos + 2)
 			goto skip;
 
-		len =	  ((p[pos + 0] & 0xff) <<  8)
-			| ((p[pos + 1] & 0xff) <<  0);
+		len = ((p[pos + 0] & 0xff) <<  8) | ((p[pos + 1] & 0xff) <<  0);
 		pos += 2;
-	}
-	else if (len == 127)
-	{
-		if (size < pos + 8)
+	} else if (len == 127) {
+		/* 8 bytes store the payload size */
+		if (size < pos + 8) {
 			goto skip;
+		}
 
 		/* Only decoding the last four bytes of the length...
 		   This limits the size of WebSocket messages that can be
 		   handled to 2^32 - which should be plenty for SIP! */
-		len =	  ((p[pos + 4] & 0xff) << 24)
+		if((p[pos] & 0xff)!=0 || (p[pos + 1] & 0xff)!=0
+				|| (p[pos + 2] & 0xff)!=0 || (p[pos + 3] & 0xff)!=0) {
+			LM_WARN("advertised length is too large (more than 2^32)\n");
+			goto skip;
+		}
+		len = ((p[pos + 4] & 0xff) << 24)
 			| ((p[pos + 5] & 0xff) << 16)
 			| ((p[pos + 6] & 0xff) <<  8)
 			| ((p[pos + 7] & 0xff) <<  0);
@@ -1154,13 +1208,18 @@ static int tcp_read_ws(struct tcp_connection *c, int* read_flags)
 	}
 
 	/* Skip mask */
-	if (mask_present)
-	{
+	if (mask_present) {
 		if (size < pos + 4)
 			goto skip;
 		pos += 4;
 	}
 
+	/* check if advertised length fits in read buffer */
+	if(len>=r->b_size) {
+		LM_WARN("advertised length (%u) greater than buffer size (%u)\n",
+				len, r->b_size);
+		goto skip;
+	}
 	/* Now check the whole message has been received */
 	if (size < pos + len)
 		goto skip;
@@ -1178,6 +1237,7 @@ static int ws_process_msg(char* tcpbuf, unsigned int len,
 {
 	int ret;
 	tcp_event_info_t tev;
+	sr_event_param_t evp = {0};
 
 	ret = 0;
 	LM_DBG("WebSocket Message: [[>>>\n%.*s<<<]]\n", len, tcpbuf);
@@ -1188,7 +1248,8 @@ static int ws_process_msg(char* tcpbuf, unsigned int len,
 		tev.len = len;
 		tev.rcv = rcv_info;
 		tev.con = con;
-		ret = sr_event_exec(SREV_TCP_WS_FRAME_IN, (void*)(&tev));
+		evp.data = (void*)(&tev);
+		ret = sr_event_exec(SREV_TCP_WS_FRAME_IN, &evp);
 	} else {
 		LM_DBG("no callback registering for handling WebSockets - dropping!\n");
 	}
@@ -1196,6 +1257,112 @@ static int ws_process_msg(char* tcpbuf, unsigned int len,
 }
 #endif
 
+static int tcp_read_hep3(struct tcp_connection *c, int* read_flags)
+{
+	int bytes;
+	uint32_t size, len;
+	char *p;
+	struct tcp_req *r;
+
+	r=&c->req;
+#ifdef USE_TLS
+	if (unlikely(c->type == PROTO_TLS)) {
+		bytes = tls_read(c, read_flags);
+	} else {
+#endif
+		bytes = tcp_read(c, read_flags);
+#ifdef USE_TLS
+	}
+#endif
+
+	if (bytes <= 0) {
+		if (likely(r->parsed >= r->pos)) {
+			LM_DBG("no new bytes to read, but still unparsed content\n");
+			return 0;
+		}
+	}
+
+	size = r->pos - r->parsed;
+
+	p = r->parsed;
+
+	/* Process first six bytes (HEP3 + 2 bytes the size)*/
+	if (size < 6) {
+		LM_DBG("not enough bytes to parse (%u)\n", size);
+		goto skip;
+	}
+
+	if(p[0]!='H' || p[1]!='E' || p[2]!='P' || p[3]!='3') {
+		/* not hep3 */
+		LM_DBG("not HEP3 packet header (%u): %c %c %c %c / %x %x %x %x\n",
+				size, p[0], p[1], p[2], p[3], p[0], p[1], p[2], p[3]);
+		goto skip;
+	}
+	r->flags |= F_TCP_REQ_HEP3;
+
+	len = ((uint32_t)(p[4] & 0xff) <<  8) + (p[5] & 0xff);
+
+	/* check if advertised length fits in read buffer */
+	if(len>=r->b_size) {
+		LM_WARN("advertised length (%u) greater than buffer size (%u)\n",
+				len, r->b_size);
+		goto skip;
+	}
+	/* check the whole message has been received */
+	if (size < len) {
+		LM_DBG("incomplete HEP3 packet (%u / %u)\n", len, size);
+		goto skip;
+	}
+
+	r->flags |= F_TCP_REQ_COMPLETE;
+	r->parsed = p + len;
+	LM_DBG("reading of HEP3 packet is complete (%u / %u)\n", len, size);
+
+skip:
+	return bytes;
+}
+
+static int hep3_process_msg(char* tcpbuf, unsigned int len,
+		struct receive_info* rcv_info, struct tcp_connection* con)
+{
+	sip_msg_t msg;
+	int ret;
+	sr_event_param_t evp = {0};
+
+	memset(&msg, 0, sizeof(sip_msg_t)); /* init everything to 0 */
+	/* fill in msg */
+	msg.buf=tcpbuf;
+	msg.len=len;
+	/* zero termination (termination of orig message below not that
+	 * useful as most of the work is done with scratch-pad; -jiri  */
+	/* buf[len]=0; */ /* WARNING: zero term removed! */
+	msg.rcv=*rcv_info;
+	msg.id=msg_no;
+	msg.pid=my_pid();
+	msg.set_global_address=default_global_address;
+	msg.set_global_port=default_global_port;
+
+	if(likely(sr_msg_time==1)) msg_set_time(&msg);
+	evp.data = (void*)(&msg);
+	ret=sr_event_exec(SREV_RCV_NOSIP, &evp);
+	LM_DBG("running hep3 handling event returned %d\n", ret);
+	if(ret == NONSIP_MSG_DROP) {
+		free_sip_msg(&msg);
+		return 0;
+	}
+	if(ret < 0) {
+		LM_ERR("error running hep3 handling event: %d\n", ret);
+		free_sip_msg(&msg);
+		return -1;
+	}
+
+	ret = receive_msg(msg.buf, msg.len, &msg.rcv);
+	LM_DBG("running hep3-enclosed sip request route returned %d\n", ret);
+	free_sip_msg(&msg);
+
+	return ret;
+}
+
 /**
  * @brief wrapper around receive_msg() to clone the tcpbuf content
  *
@@ -1208,12 +1375,8 @@ int receive_tcp_msg(char* tcpbuf, unsigned int len,
 		struct receive_info* rcv_info, struct tcp_connection* con)
 {
 #ifdef TCP_CLONE_RCVBUF
-#ifdef DYN_BUF
-	char *buf = NULL;
-#else
 	static char *buf = NULL;
 	static unsigned int bsize = 0;
-#endif
 	int blen;
 
 	/* cloning is disabled via parameter */
@@ -1226,6 +1389,8 @@ int receive_tcp_msg(char* tcpbuf, unsigned int len,
 		if(unlikely(con->type == PROTO_WS || con->type == PROTO_WSS))
 			return ws_process_msg(tcpbuf, len, rcv_info, con);
 #endif
+		if(unlikely(con->req.flags&F_TCP_REQ_HEP3))
+			return hep3_process_msg(tcpbuf, len, rcv_info, con);
 
 		return receive_msg(tcpbuf, len, rcv_info);
 	}
@@ -1235,13 +1400,6 @@ int receive_tcp_msg(char* tcpbuf, unsigned int len,
 	if(blen < BUF_SIZE)
 		blen = BUF_SIZE;
 
-#ifdef DYN_BUF
-	buf=pkg_malloc(blen+1);
-	if (buf==0) {
-		LM_ERR("could not allocate receive buffer\n");
-		return -1;
-	}
-#else
 	/* allocate buffer when needed
 	 * - no buffer yet
 	 * - existing buffer too small (min size is BUF_SIZE - to accomodate most
@@ -1257,12 +1415,11 @@ int receive_tcp_msg(char* tcpbuf, unsigned int len,
 			free(buf);
 		buf=malloc(blen+1);
 		if (buf==0) {
-			LM_ERR("could not allocate receive buffer\n");
+			SYS_MEM_ERROR;
 			return -1;
 		}
 		bsize = blen;
 	}
-#endif
 
 	memcpy(buf, tcpbuf, len);
 	buf[len] = '\0';
@@ -1274,6 +1431,8 @@ int receive_tcp_msg(char* tcpbuf, unsigned int len,
 	if(unlikely(con->type == PROTO_WS || con->type == PROTO_WSS))
 		return ws_process_msg(buf, len, rcv_info, con);
 #endif
+	if(unlikely(con->req.flags&F_TCP_REQ_HEP3))
+		return hep3_process_msg(tcpbuf, len, rcv_info, con);
 	return receive_msg(buf, len, rcv_info);
 #else /* TCP_CLONE_RCVBUF */
 #ifdef READ_MSRP
@@ -1284,6 +1443,8 @@ int receive_tcp_msg(char* tcpbuf, unsigned int len,
 	if(unlikely(con->type == PROTO_WS || con->type == PROTO_WSS))
 		return ws_process_msg(tcpbuf, len, rcv_info, con);
 #endif
+	if(unlikely(con->req.flags&F_TCP_REQ_HEP3))
+		return hep3_process_msg(tcpbuf, len, rcv_info, con);
 	return receive_msg(tcpbuf, len, rcv_info);
 #endif /* TCP_CLONE_RCVBUF */
 }
@@ -1299,24 +1460,37 @@ int tcp_read_req(struct tcp_connection* con, int* bytes_read, int* read_flags)
 	char c;
 	int ret;
 
-		bytes=-1;
-		total_bytes=0;
-		resp=CONN_RELEASE;
-		req=&con->req;
+	bytes=-1;
+	total_bytes=0;
+	resp=CONN_RELEASE;
+	req=&con->req;
 
 again:
 		if (likely(req->error==TCP_REQ_OK)){
 #ifdef READ_WS
-			if (unlikely(con->type == PROTO_WS || con->type == PROTO_WSS))
+			if (unlikely(con->type == PROTO_WS || con->type == PROTO_WSS)) {
 				bytes=tcp_read_ws(con, read_flags);
-			else
+			} else {
+#endif
+				if(unlikely(ksr_tcp_accept_hep3!=0)) {
+					bytes=tcp_read_hep3(con, read_flags);
+					if (bytes>=0) {
+						if(!(con->req.flags & F_TCP_REQ_HEP3)) {
+							/* not hep3, try to read headers */
+							bytes=tcp_read_headers(con, read_flags);
+						}
+					}
+				} else {
+					bytes=tcp_read_headers(con, read_flags);
+				}
+#ifdef READ_WS
+			}
 #endif
-				bytes=tcp_read_headers(con, read_flags);
 
-			if (unlikely(bytes==-1)){
+			if (unlikely(bytes<0)){
 				LOG(cfg_get(core, core_cfg, corelog),
-						"ERROR: tcp_read_req: error reading - c: %p r: %p\n",
-						con, req);
+						"ERROR: tcp_read_req: error reading - c: %p r: %p (%d)\n",
+						con, req, bytes);
 				resp=CONN_ERROR;
 				goto end_req;
 			}
@@ -1444,14 +1618,14 @@ again:
 #endif
 				ret = receive_tcp_msg(req->start, req->parsed-req->start,
 									&con->rcv, con);
-				
+
 			if (unlikely(ret < 0)) {
 				*req->parsed=c;
 				resp=CONN_ERROR;
 				goto end_req;
 			}
 			*req->parsed=c;
-			
+
 			/* prepare for next request */
 			size=req->pos-req->parsed;
 			req->start=req->buf;
@@ -1462,8 +1636,8 @@ again:
 			req->content_len=0;
 			req->bytes_to_go=0;
 			req->pos=req->buf+size;
-			
-			if (unlikely(size)){ 
+
+			if (unlikely(size)){
 				memmove(req->buf, req->parsed, size);
 				req->parsed=req->buf; /* fix req->parsed after using it */
 #ifdef EXTRA_DEBUG
@@ -1479,8 +1653,7 @@ again:
 			}
 			req->parsed=req->buf; /* fix req->parsed */
 		}
-		
-		
+
 	end_req:
 		if (likely(bytes_read)) *bytes_read=total_bytes;
 		return resp;
@@ -1516,10 +1689,10 @@ void release_tcpconn(struct tcp_connection* c, long state, int unix_sock)
 static ticks_t tcpconn_read_timeout(ticks_t t, struct timer_ln* tl, void* data)
 {
 	struct tcp_connection *c;
-	
-	c=(struct tcp_connection*)data; 
+
+	c=(struct tcp_connection*)data;
 	/* or (struct tcp...*)(tl-offset(c->timer)) */
-	
+
 	if (likely(!(c->state<0) && TICKS_LT(t, c->timeout))){
 		/* timeout extended, exit */
 		return (ticks_t)(c->timeout - t);
@@ -1533,9 +1706,10 @@ static ticks_t tcpconn_read_timeout(ticks_t t, struct timer_ln* tl, void* data)
 					ip_addr2a(&c->rcv.src_ip), c->rcv.src_port,
 					ip_addr2a(&c->rcv.dst_ip), c->rcv.dst_port);
 	}
-	tcpconn_listrm(tcp_conn_lst, c, c_next, c_prev);
-	release_tcpconn(c, (c->state<0)?CONN_ERROR:CONN_RELEASE, tcpmain_sock);
-	
+	if(tcp_conn_lst!=NULL) {
+		tcpconn_listrm(tcp_conn_lst, c, c_next, c_prev);
+		release_tcpconn(c, (c->state<0)?CONN_ERROR:CONN_RELEASE, tcpmain_sock);
+	}
 	return 0;
 }
 
@@ -1547,7 +1721,7 @@ static ticks_t tcpconn_read_timeout(ticks_t t, struct timer_ln* tl, void* data)
  *          idx - index in the fd_array (or -1 if not known)
  * return: -1 on error, or when we are not interested any more on reads
  *            from this fd (e.g.: we are closing it )
- *          0 on EAGAIN or when by some other way it is known that no more 
+ *          0 on EAGAIN or when by some other way it is known that no more
  *            io events are queued on the fd (the receive buffer is empty).
  *            Usefull to detect when there are no more io events queued for
  *            sigio_rt, epoll_et, kqueue.
@@ -1555,7 +1729,7 @@ static ticks_t tcpconn_read_timeout(ticks_t t, struct timer_ln* tl, void* data)
  *            queued -- the receive buffer might still be non-empty)
  */
 inline static int handle_io(struct fd_map* fm, short events, int idx)
-{	
+{
 	int ret;
 	int n;
 	int read_flags;
@@ -1563,10 +1737,11 @@ inline static int handle_io(struct fd_map* fm, short events, int idx)
 	int s;
 	long resp;
 	ticks_t t;
-	
+	fd_map_t *ee = NULL;
+
 	/* update the local config */
 	cfg_update();
-	
+
 	switch(fm->type){
 		case F_TCPMAIN:
 again:
@@ -1630,7 +1805,7 @@ repeat_1st_read:
 			if (unlikely(read_flags & RD_CONN_REPEAT_READ))
 				goto repeat_1st_read;
 #endif /* USE_TLS */
-			
+
 			/* must be before io_watch_add, io_watch_add might catch some
 			 * already existing events => might call handle_io and
 			 * handle_io might decide to del. the new connection =>
@@ -1643,15 +1818,29 @@ repeat_1st_read:
 			local_timer_reinit(&con->timer);
 			local_timer_add(&tcp_reader_ltimer, &con->timer,
 								S_TO_TICKS(TCP_CHILD_TIMEOUT), t);
-			if (unlikely(io_watch_add(&io_w, s, POLLIN, F_TCPCONN, con)<0)){
+			if (unlikely(io_watch_add(&io_w, s, POLLIN, F_TCPCONN, con)<0)) {
 				LM_CRIT("io_watch_add failed for %p id %d fd %d, state %d, flags %x,"
 							" main fd %d, refcnt %d ([%s]:%u -> [%s]:%u)\n",
 							con, con->id, con->fd, con->state, con->flags,
 							con->s, atomic_get(&con->refcnt),
 							ip_addr2a(&con->rcv.src_ip), con->rcv.src_port,
 							ip_addr2a(&con->rcv.dst_ip), con->rcv.dst_port);
-				tcpconn_listrm(tcp_conn_lst, con, c_next, c_prev);
-				local_timer_del(&tcp_reader_ltimer, &con->timer);
+				ee = get_fd_map(&io_w, s);
+				if(ee!=0 && ee->type==F_TCPCONN) {
+					tcp_connection_t *ec;
+					ec = (tcp_connection_t*)ee->data;
+					LM_CRIT("existing tcp con %p id %d fd %d, state %d, flags %x,"
+							" main fd %d, refcnt %d ([%s]:%u -> [%s]:%u)\n",
+							ec, ec->id, ec->fd, ec->state, ec->flags,
+							ec->s, atomic_get(&ec->refcnt),
+							ip_addr2a(&ec->rcv.src_ip), ec->rcv.src_port,
+							ip_addr2a(&ec->rcv.dst_ip), ec->rcv.dst_port);
+
+				}
+				if(tcp_conn_lst!=NULL) {
+					tcpconn_listrm(tcp_conn_lst, con, c_next, c_prev);
+					local_timer_del(&tcp_reader_ltimer, &con->timer);
+				}
 				goto con_error;
 			}
 			break;
@@ -1660,8 +1849,11 @@ repeat_1st_read:
 			if (unlikely(con->state==S_CONN_BAD)){
 				resp=CONN_ERROR;
 				if (!(con->send_flags.f & SND_F_CON_CLOSE))
-					LM_WARN("F_TCPCONN connection marked as bad: %p id %d refcnt %d\n",
-							con, con->id, atomic_get(&con->refcnt));
+					LM_WARN("F_TCPCONN connection marked as bad: %p id %d fd %d"
+							" refcnt %d ([%s]:%u -> [%s]:%u)\n",
+							con, con->id, con->fd, atomic_get(&con->refcnt),
+							ip_addr2a(&con->rcv.src_ip), con->rcv.src_port,
+							ip_addr2a(&con->rcv.dst_ip), con->rcv.dst_port);
 				goto read_error;
 			}
 			read_flags=((
@@ -1688,11 +1880,20 @@ read_error:
 							ip_addr2a(&con->rcv.src_ip), con->rcv.src_port,
 							ip_addr2a(&con->rcv.dst_ip), con->rcv.dst_port);
 				}
-				tcpconn_listrm(tcp_conn_lst, con, c_next, c_prev);
-				local_timer_del(&tcp_reader_ltimer, &con->timer);
-				if (unlikely(resp!=CONN_EOF))
-					con->state=S_CONN_BAD;
-				release_tcpconn(con, resp, tcpmain_sock);
+				if(tcp_conn_lst!=NULL) {
+					LM_DBG("removing from list %p id %d fd %d,"
+							" state %d, flags %x, main fd %d, refcnt %d"
+							" ([%s]:%u -> [%s]:%u)\n",
+							con, con->id, con->fd, con->state,
+							con->flags, con->s, atomic_get(&con->refcnt),
+							ip_addr2a(&con->rcv.src_ip), con->rcv.src_port,
+							ip_addr2a(&con->rcv.dst_ip), con->rcv.dst_port);
+					tcpconn_listrm(tcp_conn_lst, con, c_next, c_prev);
+					local_timer_del(&tcp_reader_ltimer, &con->timer);
+					if (unlikely(resp!=CONN_EOF))
+						con->state=S_CONN_BAD;
+					release_tcpconn(con, resp, tcpmain_sock);
+				}
 			}else{
 #ifdef USE_TLS
 				if (unlikely(read_flags & RD_CONN_REPEAT_READ))
@@ -1700,8 +1901,8 @@ read_error:
 #endif /* USE_TLS */
 				/* update timeout */
 				con->timeout=get_ticks_raw()+S_TO_TICKS(TCP_CHILD_TIMEOUT);
-				/* ret= 0 (read the whole socket buffer) if short read & 
-				 *  !POLLPRI,  bytes read otherwise */
+				/* ret= 0 (read the whole socket buffer) if short read
+				 * & !POLLPRI,  bytes read otherwise */
 				ret&=(((read_flags & RD_CONN_SHORT_READ) &&
 						!(events & POLLPRI)) - 1);
 			}
@@ -1712,10 +1913,10 @@ read_error:
 						fm->fd, fm->type, fm->data);
 			goto error;
 		default:
-			LM_CRIT("uknown fd type %d\n", fm->type); 
+			LM_CRIT("unknown fd type %d\n", fm->type);
 			goto error;
 	}
-	
+
 	return ret;
 con_error:
 	con->state=S_CONN_BAD;
