@@ -39,6 +39,7 @@
 #include "../../core/dset.h"
 #include "../../core/basex.h"
 #include "../../core/action.h"
+#include "../../core/hashes.h"
 
 #include "../../core/parser/parse_param.h"
 #include "../../core/parser/parse_uri.h"
@@ -46,7 +47,7 @@
 #include "../../core/parser/parse_nameaddr.h"
 
 #include "../../core/strutils.h"
-#include "../../lib/srutils/shautils.h"
+#include "../../core/crypto/shautils.h"
 #include "pv_trans.h"
 
 
@@ -58,7 +59,7 @@ static param_t* _tr_uri_params = NULL;
 
 /*! transformation buffer size */
 #define TR_BUFFER_SIZE 65536
-#define TR_BUFFER_SLOTS	4
+#define TR_BUFFER_SLOTS	8
 
 /*! transformation buffer */
 static char **_tr_buffer_list = NULL;
@@ -106,78 +107,6 @@ char *tr_set_crt_buffer(void)
 
 /* -- helper functions */
 
-/* Converts a hex character to its integer value */
-static char pv_from_hex(char ch)
-{
-	return isdigit(ch) ? ch - '0' : tolower(ch) - 'a' + 10;
-}
-
-/* Converts an integer value to its hex character */
-static char pv_to_hex(char code)
-{
-	static char hex[] = "0123456789abcdef";
-	return hex[code & 15];
-}
-
-/*! \brief
- *  URL Encodes a string for use in a HTTP query
- */
-static int urlencode_param(str *sin, str *sout)
-{
-	char *at, *p;
-
-	if (sin==NULL || sout==NULL || sin->s==NULL || sout->s==NULL ||
-			sin->len<0 || sout->len < 3*sin->len+1)
-		return -1;
-
-	at = sout->s;
-	p  = sin->s;
-
-	while (p < sin->s+sin->len) {
-		if (isalnum(*p) || *p == '-' || *p == '_' || *p == '.' || *p == '~')
-			*at++ = *p;
-		else if (*p == ' ')
-			*at++ = '+';
-		else
-			*at++ = '%', *at++ = pv_to_hex(*p >> 4), *at++ = pv_to_hex(*p & 15);
-		p++;
-	}
-
-	*at = 0;
-	sout->len = at - sout->s;
-	LM_DBG("urlencoded string is <%s>\n", sout->s);
-
-	return 0;
-}
-
-/* URL Decode a string */
-static int urldecode_param(str *sin, str *sout) {
-	char *at, *p;
-
-	at = sout->s;
-	p  = sin->s;
-
-	while (p < sin->s+sin->len) {
-		if (*p == '%') {
-			if (p[1] && p[2]) {
-				*at++ = pv_from_hex(p[1]) << 4 | pv_from_hex(p[2]);
-				p += 2;
-			}
-		} else if (*p == '+') {
-			*at++ = ' ';
-		} else {
-			*at++ = *p;
-		}
-		p++;
-	}
-
-	*at = 0;
-	sout->len = at - sout->s;
-
-	LM_DBG("urldecoded string is <%s>\n", sout->s);
-	return 0;
-}
-
 /* Encode 7BIT PDU */
 static int pdu_7bit_encode(str sin) {
 	int i, j;
@@ -214,8 +143,8 @@ static int pdu_7bit_decode(str sin) {
 	unsigned char oldfill = 0;
 	j = 0;
 	for(i = 0; i < sin.len; i += 2) {
-		_tr_buffer[j] = (unsigned char)pv_from_hex(sin.s[i]) << 4;
-		_tr_buffer[j] |= (unsigned char)pv_from_hex(sin.s[i+1]);
+		_tr_buffer[j] = (unsigned char)hex_to_char(sin.s[i]) << 4;
+		_tr_buffer[j] |= (unsigned char)hex_to_char(sin.s[i+1]);
 		fill = (unsigned char)_tr_buffer[j];
 		fill >>= (8 - nleft);
 		_tr_buffer[j] <<= (nleft -1 );
@@ -265,6 +194,8 @@ int tr_eval_string(struct sip_msg *msg, tr_param_t *tp, int subtype,
 	str st, st2;
 	pv_value_t v, w;
 	time_t t;
+	uint32_t sz1, sz2;
+	struct tm tmv;
 
 	if(val==NULL || val->flags&PV_VAL_NULL)
 		return -1;
@@ -293,6 +224,25 @@ int tr_eval_string(struct sip_msg *msg, tr_param_t *tp, int subtype,
 
 			val->flags = PV_TYPE_INT|PV_VAL_INT|PV_VAL_STR;
 			break;
+		case TR_S_RMWS:
+			if(!(val->flags&PV_VAL_STR))
+				val->rs.s = int2str(val->ri, &val->rs.len);
+			if(val->rs.len >= TR_BUFFER_SIZE - 1)
+				return -1;
+			j = 0;
+			for(i=0; i < val->rs.len; i++) {
+				if(val->rs.s[i] != ' ' && val->rs.s[i] != '\t'
+						&& val->rs.s[i] != '\r' && val->rs.s[i] != '\n') {
+					_tr_buffer[j] = val->rs.s[i];
+					j++;
+				}
+			}
+			_tr_buffer[j] = '\0';
+			val->flags = PV_VAL_STR;
+			val->ri = 0;
+			val->rs.s = _tr_buffer;
+			val->rs.len = j;
+			break;
 		case TR_S_MD5:
 			if(!(val->flags&PV_VAL_STR))
 				val->rs.s = int2str(val->ri, &val->rs.len);
@@ -413,6 +363,30 @@ int tr_eval_string(struct sip_msg *msg, tr_param_t *tp, int subtype,
 			val->rs.s = _tr_buffer;
 			val->rs.len = i;
 			break;
+		case TR_S_ENCODEBASE58:
+			if(!(val->flags&PV_VAL_STR))
+				val->rs.s = int2str(val->ri, &val->rs.len);
+			st.len = TR_BUFFER_SIZE-1;
+			st.s = b58_encode(_tr_buffer, &st.len, val->rs.s, val->rs.len);
+			if (st.s==NULL)
+				return -1;
+			memset(val, 0, sizeof(pv_value_t));
+			val->flags = PV_VAL_STR;
+			val->rs.s = st.s;
+			val->rs.len = st.len;
+			break;
+		case TR_S_DECODEBASE58:
+			if(!(val->flags&PV_VAL_STR))
+				val->rs.s = int2str(val->ri, &val->rs.len);
+			st.len = TR_BUFFER_SIZE-1;
+			st.s = b58_decode(_tr_buffer, &st.len, val->rs.s, val->rs.len);
+			if (st.s==NULL)
+				return -1;
+			memset(val, 0, sizeof(pv_value_t));
+			val->flags = PV_VAL_STR;
+			val->rs.s = st.s;
+			val->rs.len = st.len;
+			break;
 		case TR_S_ENCODEBASE64:
 			if(!(val->flags&PV_VAL_STR))
 				val->rs.s = int2str(val->ri, &val->rs.len);
@@ -439,6 +413,130 @@ int tr_eval_string(struct sip_msg *msg, tr_param_t *tp, int subtype,
 			val->rs.s = _tr_buffer;
 			val->rs.len = i;
 			break;
+		case TR_S_ENCODEBASE64T:
+			if(!(val->flags&PV_VAL_STR))
+				val->rs.s = int2str(val->ri, &val->rs.len);
+			i = base64_enc((unsigned char *) val->rs.s, val->rs.len,
+					(unsigned char *) _tr_buffer, TR_BUFFER_SIZE-1);
+			if (i < 0)
+				return -1;
+			if (i>1 && _tr_buffer[i-1] == '=') {
+				i--;
+				if (i>1 && _tr_buffer[i-1] == '=') {
+					i--;
+				}
+			}
+			_tr_buffer[i] = '\0';
+			memset(val, 0, sizeof(pv_value_t));
+			val->flags = PV_VAL_STR;
+			val->rs.s = _tr_buffer;
+			val->rs.len = i;
+			break;
+		case TR_S_DECODEBASE64T:
+			if(!(val->flags&PV_VAL_STR))
+				val->rs.s = int2str(val->ri, &val->rs.len);
+			if(val->rs.len % 4) {
+				if(val->rs.len + 4 >= TR_BUFFER_SIZE-1) {
+					LM_ERR("not enough space to insert padding\n");
+					return -1;
+				}
+				memcpy(_tr_buffer, val->rs.s, val->rs.len);
+				for(i=0; i < (4 - (val->rs.len % 4)); i++) {
+					_tr_buffer[val->rs.len + i] = '=';
+				}
+				st.s = _tr_buffer;
+				st.len = val->rs.len + (4 - (val->rs.len % 4));
+				/* move to next buffer */
+				tr_set_crt_buffer();
+				i = base64_dec((unsigned char *) st.s, st.len,
+						(unsigned char *) _tr_buffer, TR_BUFFER_SIZE-1);
+			} else {
+				i = base64_dec((unsigned char *) val->rs.s, val->rs.len,
+						(unsigned char *) _tr_buffer, TR_BUFFER_SIZE-1);
+			}
+			if (i < 0 || (i == 0 && val->rs.len > 0))
+				return -1;
+			_tr_buffer[i] = '\0';
+			memset(val, 0, sizeof(pv_value_t));
+			val->flags = PV_VAL_STR;
+			val->rs.s = _tr_buffer;
+			val->rs.len = i;
+			break;
+		case TR_S_ENCODEBASE64URL:
+			if(!(val->flags&PV_VAL_STR))
+				val->rs.s = int2str(val->ri, &val->rs.len);
+			i = base64url_enc(val->rs.s, val->rs.len,
+					_tr_buffer, TR_BUFFER_SIZE-1);
+			if (i < 0)
+				return -1;
+			_tr_buffer[i] = '\0';
+			memset(val, 0, sizeof(pv_value_t));
+			val->flags = PV_VAL_STR;
+			val->rs.s = _tr_buffer;
+			val->rs.len = i;
+			break;
+		case TR_S_DECODEBASE64URL:
+			if(!(val->flags&PV_VAL_STR))
+				val->rs.s = int2str(val->ri, &val->rs.len);
+			i = base64url_dec(val->rs.s, val->rs.len,
+					_tr_buffer, TR_BUFFER_SIZE-1);
+			if (i < 0 || (i == 0 && val->rs.len > 0))
+				return -1;
+			_tr_buffer[i] = '\0';
+			memset(val, 0, sizeof(pv_value_t));
+			val->flags = PV_VAL_STR;
+			val->rs.s = _tr_buffer;
+			val->rs.len = i;
+			break;
+		case TR_S_ENCODEBASE64URLT:
+			if(!(val->flags&PV_VAL_STR))
+				val->rs.s = int2str(val->ri, &val->rs.len);
+			i = base64url_enc(val->rs.s, val->rs.len,
+					_tr_buffer, TR_BUFFER_SIZE-1);
+			if (i < 0)
+				return -1;
+			if (i>1 && _tr_buffer[i-1] == '=') {
+				i--;
+				if (i>1 && _tr_buffer[i-1] == '=') {
+					i--;
+				}
+			}
+			_tr_buffer[i] = '\0';
+			memset(val, 0, sizeof(pv_value_t));
+			val->flags = PV_VAL_STR;
+			val->rs.s = _tr_buffer;
+			val->rs.len = i;
+			break;
+		case TR_S_DECODEBASE64URLT:
+			if(!(val->flags&PV_VAL_STR))
+				val->rs.s = int2str(val->ri, &val->rs.len);
+			if(val->rs.len % 4) {
+				if(val->rs.len + 4 >= TR_BUFFER_SIZE-1) {
+					LM_ERR("not enough space to insert padding\n");
+					return -1;
+				}
+				memcpy(_tr_buffer, val->rs.s, val->rs.len);
+				for(i=0; i < (4 - (val->rs.len % 4)); i++) {
+					_tr_buffer[val->rs.len + i] = '=';
+				}
+				st.s = _tr_buffer;
+				st.len = val->rs.len + (4 - (val->rs.len % 4));
+				/* move to next buffer */
+				tr_set_crt_buffer();
+				i = base64url_dec(st.s, st.len,
+						_tr_buffer, TR_BUFFER_SIZE-1);
+			} else {
+				i = base64url_dec(val->rs.s, val->rs.len,
+						_tr_buffer, TR_BUFFER_SIZE-1);
+			}
+			if (i < 0 || (i == 0 && val->rs.len > 0))
+				return -1;
+			_tr_buffer[i] = '\0';
+			memset(val, 0, sizeof(pv_value_t));
+			val->flags = PV_VAL_STR;
+			val->rs.s = _tr_buffer;
+			val->rs.len = i;
+			break;
 		case TR_S_ESCAPECOMMON:
 			if(!(val->flags&PV_VAL_STR))
 				val->rs.s = int2str(val->ri, &val->rs.len);
@@ -588,6 +686,7 @@ int tr_eval_string(struct sip_msg *msg, tr_param_t *tp, int subtype,
 				}
 				val->rs.s += i;
 				val->rs.len = j;
+				tr_string_clone_result;
 				break;
 			}
 			i = -i;
@@ -810,6 +909,7 @@ int tr_eval_string(struct sip_msg *msg, tr_param_t *tp, int subtype,
 			if(!(val->flags&PV_VAL_STR))
 				val->rs.s = int2str(val->ri, &val->rs.len);
 
+
 			/* Set maximum prefix length */
 			max = val->rs.len;
 			if(tp!=NULL) {
@@ -848,7 +948,7 @@ int tr_eval_string(struct sip_msg *msg, tr_param_t *tp, int subtype,
 			memset(val, 0, sizeof(pv_value_t));
 			val->flags = PV_VAL_STR;
 			val->rs.s = _tr_buffer;
-			val->rs.len = j-1;
+			val->rs.len = (j>0)?(j-1):0;
 			break;
 
 
@@ -947,8 +1047,8 @@ int tr_eval_string(struct sip_msg *msg, tr_param_t *tp, int subtype,
 			memcpy(s, st.s, st.len);
 			s[st.len] = '\0';
 			t = val->ri;
-			val->rs.len = strftime(_tr_buffer, TR_BUFFER_SIZE-1, s,
-					localtime(&t));
+			localtime_r(&t, &tmv);
+			val->rs.len = strftime(_tr_buffer, TR_BUFFER_SIZE-1, s, &tmv);
 			pkg_free(s);
 			val->flags = PV_VAL_STR;
 			val->rs.s = _tr_buffer;
@@ -1076,7 +1176,7 @@ int tr_eval_string(struct sip_msg *msg, tr_param_t *tp, int subtype,
 				return -1;
 			st.s = _tr_buffer;
 			st.len = TR_BUFFER_SIZE;
-			if (urlencode_param(&val->rs, &st) < 0)
+			if (urlencode(&val->rs, &st) < 0)
 				return -1;
 			memset(val, 0, sizeof(pv_value_t));
 			val->flags = PV_VAL_STR;
@@ -1090,13 +1190,132 @@ int tr_eval_string(struct sip_msg *msg, tr_param_t *tp, int subtype,
 				return -1;
 			st.s = _tr_buffer;
 			st.len = TR_BUFFER_SIZE;
-			if (urldecode_param(&val->rs, &st) < 0)
+			if (urldecode(&val->rs, &st) < 0)
 				return -1;
 			memset(val, 0, sizeof(pv_value_t));
 			val->flags = PV_VAL_STR;
 			val->rs = st;
 			break;
 
+		case TR_S_COREHASH:
+			if(!(val->flags&PV_VAL_STR))
+				st.s = int2str(val->ri, &st.len);
+			else
+				st = val->rs;
+
+			sz1 = 0;
+			if(tp != NULL) {
+				if(tp->type==TR_PARAM_NUMBER) {
+						sz1 = (uint) tp->v.n;
+				} else {
+					if(pv_get_spec_value(msg, (pv_spec_p)tp->v.data, &v)!=0
+							|| (!(v.flags&PV_VAL_INT)))
+					{
+							LM_ERR("corehash cannot get size (cfg line: %d)\n",
+								get_cfg_crt_line());
+						return -1;
+					}
+					sz1  = (uint) v.ri;
+				}
+			}
+
+			sz2 = core_hash(&st, NULL, sz1);
+
+			if((val->rs.s = int2strbuf((unsigned long)sz2, _tr_buffer,
+						INT2STR_MAX_LEN, &val->rs.len))==NULL) {
+				LM_ERR("failed to convert core hash id to string\n");
+				return -1;
+			}
+			val->flags = PV_VAL_STR;
+			val->ri = 0;
+			break;
+
+		case TR_S_UNQUOTE:
+			if(!(val->flags&PV_VAL_STR)) {
+				val->rs.s = int2str(val->ri, &val->rs.len);
+				break;
+			}
+			if(val->rs.len<2) {
+				break;
+			}
+			if(val->rs.len>TR_BUFFER_SIZE-2) {
+				LM_ERR("value too large: %d\n", val->rs.len);
+				return -1;
+			}
+			if((val->rs.s[0] == val->rs.s[val->rs.len-1])
+					&& (val->rs.s[0] == '"' || val->rs.s[0] == '\'')) {
+				memcpy(_tr_buffer, val->rs.s+1, val->rs.len-2);
+				val->rs.len -= 2;
+			} else {
+				memcpy(_tr_buffer, val->rs.s, val->rs.len);
+			}
+			val->flags = PV_VAL_STR;
+			val->rs.s = _tr_buffer;
+			val->rs.s[val->rs.len] = '\0';
+			break;
+
+		case TR_S_UNBRACKET:
+			if(!(val->flags&PV_VAL_STR)) {
+				val->rs.s = int2str(val->ri, &val->rs.len);
+				break;
+			}
+			if(val->rs.len<2) {
+				break;
+			}
+			if(val->rs.len>TR_BUFFER_SIZE-2) {
+				LM_ERR("value too large: %d\n", val->rs.len);
+				return -1;
+			}
+			if((val->rs.s[0] == '(' && val->rs.s[val->rs.len-1] == ')')
+					|| (val->rs.s[0] == '[' && val->rs.s[val->rs.len-1] == ']')
+					|| (val->rs.s[0] == '{' && val->rs.s[val->rs.len-1] == '}')
+					|| (val->rs.s[0] == '<' && val->rs.s[val->rs.len-1] == '>')) {
+				memcpy(_tr_buffer, val->rs.s+1, val->rs.len-2);
+				val->rs.len -= 2;
+			} else {
+				memcpy(_tr_buffer, val->rs.s, val->rs.len);
+			}
+			val->flags = PV_VAL_STR;
+			val->rs.s = _tr_buffer;
+			val->rs.s[val->rs.len] = '\0';
+			break;
+
+		case TR_S_COUNT:
+			if(tp==NULL)
+			{
+				LM_ERR("invalid parameters (cfg line: %d)\n",
+						get_cfg_crt_line());
+				return -1;
+			}
+			if(!(val->flags&PV_VAL_STR)) {
+				val->rs.s = int2str(val->ri, &val->rs.len);
+			}
+			if(tp->type==TR_PARAM_STRING)
+			{
+				st = tp->v.s;
+			} else {
+				if(pv_get_spec_value(msg, (pv_spec_p)tp->v.data, &v)!=0
+						|| (!(v.flags&PV_VAL_STR)) || v.rs.len<=0)
+				{
+					LM_ERR("cannot get parameter value (cfg line: %d)\n",
+							get_cfg_crt_line());
+					return -1;
+				}
+				st = v.rs;
+			}
+			LM_DBG("counting [%.*s](%d) in [%.*s](%d)\n",
+					st.len, st.s, st.len, val->rs.len, val->rs.s, val->rs.len);
+			j = 0;
+			for(i=0; i<val->rs.len; i++) {
+				if(val->rs.s[i]==st.s[0]) {
+					j++;
+				}
+			}
+			val->flags = PV_TYPE_INT|PV_VAL_INT|PV_VAL_STR;
+			val->ri = j;
+			val->rs.s = int2str(j, &val->rs.len);
+			break;
+
 		default:
 			LM_ERR("unknown subtype %d (cfg line: %d)\n",
 					subtype, get_cfg_crt_line());
@@ -1121,6 +1340,8 @@ int tr_eval_uri(struct sip_msg *msg, tr_param_t *tp, int subtype,
 	str sv;
 	param_hooks_t phooks;
 	param_t *pit=NULL;
+	str sproto;
+	int dlen = 0;
 
 	if(val==NULL || (!(val->flags&PV_VAL_STR)) || val->rs.len<=0)
 		return -1;
@@ -1155,8 +1376,11 @@ int tr_eval_uri(struct sip_msg *msg, tr_param_t *tp, int subtype,
 			free_params(_tr_uri_params);
 			_tr_uri_params = 0;
 		}
+		if(_tr_uri.len>4 && _tr_uri.s[_tr_uri.len-1]==';') {
+			dlen = 1;
+		}
 		/* parse uri -- params only when requested */
-		if(parse_uri(_tr_uri.s, _tr_uri.len, &_tr_parsed_uri)!=0)
+		if(parse_uri(_tr_uri.s, _tr_uri.len - dlen, &_tr_parsed_uri)!=0)
 		{
 			LM_ERR("invalid uri [%.*s]\n", val->rs.len,
 					val->rs.s);
@@ -1268,6 +1492,36 @@ int tr_eval_uri(struct sip_msg *msg, tr_param_t *tp, int subtype,
 			val->rs = (_tr_parsed_uri.r2_val.s)?
 				_tr_parsed_uri.r2_val:_tr_empty;
 			break;
+		case TR_URI_SCHEME:
+			val->rs.s = _tr_uri.s;
+			val->rs.len = 0;
+			while(val->rs.len<_tr_uri.len) {
+				if(_tr_uri.s[val->rs.len]==':') {
+					break;
+				}
+				val->rs.len++;
+			}
+			break;
+		case TR_URI_TOSOCKET:
+			if(get_valid_proto_string(_tr_parsed_uri.proto, 1, 0, &sproto)<0) {
+				LM_WARN("unknown transport protocol\n");
+				val->rs = _tr_empty;
+				break;
+			}
+			tr_set_crt_buffer();
+			val->rs.len = snprintf(_tr_buffer, TR_BUFFER_SIZE,
+					"%.*s:%.*s:%d", sproto.len, sproto.s,
+					_tr_parsed_uri.host.len, _tr_parsed_uri.host.s,
+					(_tr_parsed_uri.port_no!=0)
+							?(int)_tr_parsed_uri.port_no:5060);
+			if(val->rs.len<=0 || val->rs.len>=TR_BUFFER_SIZE) {
+				LM_WARN("error converting uri to socket address [%.*s]\n",
+						_tr_uri.len, _tr_uri.s);
+				val->rs = _tr_empty;
+				break;
+			}
+			val->rs.s = _tr_buffer;
+			break;
 		default:
 			LM_ERR("unknown subtype %d\n",
 					subtype);
@@ -1364,6 +1618,9 @@ int tr_eval_paramlist(struct sip_msg *msg, tr_param_t *tp, int subtype,
 
 		/* parse params */
 		sv = _tr_params_str;
+		if(sv.len>1 && sv.s[sv.len - 1] == _tr_params_separator) {
+			sv.len--;
+		}
 		if (parse_params2(&sv, CLASS_ANY, &phooks, &_tr_params_list,
 					_tr_params_separator)<0)
 			return -1;
@@ -1982,28 +2239,28 @@ done:
 		} \
 	}
 
-#define _tr_parse_sparam(_p, _p0, _tp, _spec, _ps, _in, _s) \
+/**
+ * _m: 1 - the parameter value can be the TR_PARAM_MARKER; 0 - not allowed
+ */
+#define _tr_parse_sparamx(_p, _p0, _tp, _spec, _ps, _in, _s, _m) \
 	while(is_in_str(_p, _in) && (*_p==' ' || *_p=='\t' || *_p=='\n')) _p++; \
 	if(*_p==PV_MARKER) \
 	{ /* pseudo-variable */ \
 		_spec = (pv_spec_t*)pkg_malloc(sizeof(pv_spec_t)); \
-		if(_spec==NULL) \
-		{ \
+		if(_spec==NULL) { \
 			LM_ERR("no more private memory!\n"); \
 			goto error; \
 		} \
 		_s.s = _p; _s.len = _in->s + _in->len - _p; \
 		_p0 = pv_parse_spec(&_s, _spec); \
-		if(_p0==NULL) \
-		{ \
+		if(_p0==NULL) { \
 			LM_ERR("invalid spec in substr transformation: %.*s!\n", \
 					_in->len, _in->s); \
 			goto error; \
 		} \
 		_p = _p0; \
 		_tp = (tr_param_t*)pkg_malloc(sizeof(tr_param_t)); \
-		if(_tp==NULL) \
-		{ \
+		if(_tp==NULL) { \
 			LM_ERR("no more private memory!\n"); \
 			goto error; \
 		} \
@@ -2014,17 +2271,23 @@ done:
 		_ps = _p; \
 		while(is_in_str(_p, _in) && *_p!='\t' && *_p!='\n' \
 				&& *_p!=TR_PARAM_MARKER && *_p!=TR_RBRACKET) \
-		_p++; \
-		if(*_p=='\0') \
-		{ \
+			_p++; \
+		if(*_p=='\0') { \
 			LM_ERR("invalid param in transformation: %.*s!!\n", \
 					_in->len, _in->s); \
 			goto error; \
 		} \
+		if(_m && *_p==TR_PARAM_MARKER) { \
+			_p++; \
+			if(*_p=='\0') { \
+				LM_ERR("invalid param in transformation: %.*s!!!\n", \
+					_in->len, _in->s); \
+				goto error; \
+			} \
+		} \
 		_tp = (tr_param_t*)pkg_malloc(sizeof(tr_param_t)); \
-		if(_tp==NULL) \
-		{ \
-			LM_ERR("no more private memory!\n"); \
+		if(_tp==NULL) { \
+			LM_ERR("no more private memory!!\n"); \
 			goto error; \
 		} \
 		memset(_tp, 0, sizeof(tr_param_t)); \
@@ -2033,6 +2296,8 @@ done:
 		_tp->v.s.len = _p - _ps; \
 	}
 
+#define _tr_parse_sparam(_p, _p0, _tp, _spec, _ps, _in, _s) \
+	_tr_parse_sparamx(_p, _p0, _tp, _spec, _ps, _in, _s, 0)
 
 /*!
  * \brief Helper fuction to parse a string transformation
@@ -2081,6 +2346,9 @@ char* tr_parse_string(str* in, trans_t *t)
 	} else if(name.len==3 && strncasecmp(name.s, "md5", 3)==0) {
 		t->subtype = TR_S_MD5;
 		goto done;
+	} else if(name.len==4 && strncasecmp(name.s, "rmws", 4)==0) {
+		t->subtype = TR_S_RMWS;
+		goto done;
 	} else if(name.len==6 && strncasecmp(name.s, "sha256", 6)==0) {
 		t->subtype = TR_S_SHA256;
 		goto done;
@@ -2111,12 +2379,36 @@ char* tr_parse_string(str* in, trans_t *t)
 	} else if(name.len==11 && strncasecmp(name.s, "decode.7bit", 11)==0) {
 		t->subtype = TR_S_DECODE7BIT;
 		goto done;
+	} else if(name.len==13 && strncasecmp(name.s, "encode.base58", 13)==0) {
+		t->subtype = TR_S_ENCODEBASE58;
+		goto done;
+	} else if(name.len==13 && strncasecmp(name.s, "decode.base58", 13)==0) {
+		t->subtype = TR_S_DECODEBASE58;
+		goto done;
 	} else if(name.len==13 && strncasecmp(name.s, "encode.base64", 13)==0) {
 		t->subtype = TR_S_ENCODEBASE64;
 		goto done;
 	} else if(name.len==13 && strncasecmp(name.s, "decode.base64", 13)==0) {
 		t->subtype = TR_S_DECODEBASE64;
 		goto done;
+	} else if(name.len==14 && strncasecmp(name.s, "encode.base64t", 14)==0) {
+		t->subtype = TR_S_ENCODEBASE64T;
+		goto done;
+	} else if(name.len==14 && strncasecmp(name.s, "decode.base64t", 14)==0) {
+		t->subtype = TR_S_DECODEBASE64T;
+		goto done;
+	} else if(name.len==16 && strncasecmp(name.s, "encode.base64url", 16)==0) {
+		t->subtype = TR_S_ENCODEBASE64URL;
+		goto done;
+	} else if(name.len==16 && strncasecmp(name.s, "decode.base64url", 16)==0) {
+		t->subtype = TR_S_DECODEBASE64URL;
+		goto done;
+	} else if(name.len==17 && strncasecmp(name.s, "encode.base64urlt", 17)==0) {
+		t->subtype = TR_S_ENCODEBASE64URLT;
+		goto done;
+	} else if(name.len==17 && strncasecmp(name.s, "decode.base64urlt", 17)==0) {
+		t->subtype = TR_S_DECODEBASE64URLT;
+		goto done;
 	} else if(name.len==13 && strncasecmp(name.s, "escape.common", 13)==0) {
 		t->subtype = TR_S_ESCAPECOMMON;
 		goto done;
@@ -2355,6 +2647,23 @@ char* tr_parse_string(str* in, trans_t *t)
 			goto error;
 		}
 		goto done;
+	} else if(name.len==8 && strncasecmp(name.s, "corehash", 8)==0) {
+		t->subtype = TR_S_COREHASH;
+		if(*p==TR_PARAM_MARKER)
+		{
+			p++;
+			_tr_parse_nparam(p, p0, tp, spec, n, sign, in, s);
+			t->params = tp;
+		}
+		tp = 0;
+		while(*p && (*p==' ' || *p=='\t' || *p=='\n')) p++;
+		if(*p!=TR_RBRACKET)
+		{
+			LM_ERR("invalid corehash transformation: %.*s!!\n",
+					in->len, in->s);
+			goto error;
+		}
+		goto done;
 	} else if(name.len==4 && strncasecmp(name.s, "trim", 4)==0) {
 		t->subtype = TR_S_TRIM;
 		goto done;
@@ -2390,6 +2699,32 @@ char* tr_parse_string(str* in, trans_t *t)
 	} else if(name.len==15 && strncasecmp(name.s, "urldecode.param", 15)==0) {
 		t->subtype = TR_S_URLDECODEPARAM;
 		goto done;
+	} else if(name.len==7 && strncasecmp(name.s, "unquote", 7)==0) {
+		t->subtype = TR_S_UNQUOTE;
+		goto done;
+	} else if(name.len==9 && strncasecmp(name.s, "unbracket", 9)==0) {
+		t->subtype = TR_S_UNBRACKET;
+		goto done;
+	} else if(name.len==5 && strncasecmp(name.s, "count", 5)==0) {
+		t->subtype = TR_S_COUNT;
+		if(*p!=TR_PARAM_MARKER)
+		{
+			LM_ERR("invalid count transformation: %.*s!\n",
+					in->len, in->s);
+			goto error;
+		}
+		p++;
+		_tr_parse_sparamx(p, p0, tp, spec, ps, in, s, 1);
+		t->params = tp;
+		tp = 0;
+		while(*p && (*p==' ' || *p=='\t' || *p=='\n')) p++;
+		if(*p!=TR_RBRACKET)
+		{
+			LM_ERR("invalid count transformation: %.*s!!\n",
+					in->len, in->s);
+			goto error;
+		}
+		goto done;
 	}
 
 	LM_ERR("unknown transformation: %.*s/%.*s/%d!\n", in->len, in->s,
@@ -2499,16 +2834,21 @@ char* tr_parse_uri(str* in, trans_t *t)
 	} else if(name.len==2 && strncasecmp(name.s, "r2", 2)==0) {
 		t->subtype = TR_URI_R2;
 		goto done;
+	} else if(name.len==6 && strncasecmp(name.s, "scheme", 6)==0) {
+		t->subtype = TR_URI_SCHEME;
+		goto done;
+	} else if(name.len==8 && strncasecmp(name.s, "tosocket", 8)==0) {
+		t->subtype = TR_URI_TOSOCKET;
+		goto done;
 	}
 
 	LM_ERR("unknown transformation: %.*s/%.*s!\n", in->len,
 			in->s, name.len, name.s);
 error:
-	if(tp)
-		tr_param_free(tp);
 	if(spec)
 		pv_spec_free(spec);
 	return NULL;
+
 done:
 	t->name = name;
 	return p;
@@ -2664,13 +3004,11 @@ char* tr_parse_paramlist(str* in, trans_t *t)
 		goto done;
 	} else if(name.len==5 && strncasecmp(name.s, "count", 5)==0) {
 		t->subtype = TR_PL_COUNT;
-		if(*p==TR_PARAM_MARKER)
-		{
+		if(*p==TR_PARAM_MARKER) {
 			start_pos = ++p;
-			_tr_parse_sparam(p, p0, tp, spec, ps, in, s);
+			_tr_parse_sparamx(p, p0, tp, spec, ps, in, s, 1);
 			t->params = tp;
-			if (tp->type != TR_PARAM_SPEC && p - start_pos != 1)
-			{
+			if (tp->type != TR_PARAM_SPEC && p - start_pos != 1) {
 				LM_ERR("invalid separator in transformation: "
 						"%.*s\n", in->len, in->s);
 				goto error;
@@ -2678,8 +3016,7 @@ char* tr_parse_paramlist(str* in, trans_t *t)
 			tp = 0;
 
 			while(*p && (*p==' ' || *p=='\t' || *p=='\n')) p++;
-			if(*p!=TR_RBRACKET)
-			{
+			if(*p!=TR_RBRACKET) {
 				LM_ERR("invalid name transformation: %.*s!\n",
 						in->len, in->s);
 				goto error;
