@@ -13,8 +13,8 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License 
- * along with this program; if not, write to the Free Software 
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
@@ -69,9 +69,11 @@
 
 #define RR_PARAM_BUF_SIZE 512 /*!< buffer for RR parameter */
 
+extern int rr_ignore_sips;
+extern int rr_sockname_mode;
 
 /*!
- * \brief RR param buffer 
+ * \brief RR param buffer
  * \note used for storing RR param which are added before RR insertion
  */
 static char rr_param_buf_ptr[RR_PARAM_BUF_SIZE];
@@ -139,7 +141,7 @@ static inline int get_username(struct sip_msg* _m, str* _user)
 	/* no username in original uri -- hmm; maybe it is a uri
 	 * with just host address and username is in a preloaded route,
 	 * which is now no rewritten r-uri (assumed rewriteFromRoute
-	 * was called somewhere in script's beginning) 
+	 * was called somewhere in script's beginning)
 	 */
 	if (!puri.user.len && _m->new_uri.s) {
 		if (parse_uri(_m->new_uri.s, _m->new_uri.len, &puri) < 0) {
@@ -206,6 +208,7 @@ static inline int build_rr(struct lump* _l, struct lump* _l2, str* user,
 	char *p;
 	char *rr_prefix;
 	int rr_prefix_len;
+	int rr_lump_type;
 
 	if(_sips==0) {
 		rr_prefix = RR_PREFIX_SIP;
@@ -237,13 +240,13 @@ static inline int build_rr(struct lump* _l, struct lump* _l2, str* user,
 		if (r2) pkg_free(r2);
 		return -3;
 	}
-	
+
 	memcpy(prefix, rr_prefix, rr_prefix_len);
 	if (user->len) {
 		memcpy(prefix + rr_prefix_len, user->s, user->len);
 #ifdef ENABLE_USER_CHECK
 		/* don't add the ignored user into a RR */
-		if(i_user.len && i_user.len == user->len && 
+		if(i_user.len && i_user.len == user->len &&
 				!strncmp(i_user.s, user->s, i_user.len))
 		{
 			if(prefix[rr_prefix_len]=='x')
@@ -273,14 +276,17 @@ static inline int build_rr(struct lump* _l, struct lump* _l2, str* user,
 		memcpy(p, params->s, params->len);
 		p += params->len;
 	}
-	
+
 	memcpy(term, RR_TERM, RR_TERM_LEN);
 	memcpy(r2, RR_R2, RR_R2_LEN);
 
-	if (!(_l = insert_new_lump_after(_l, prefix, prefix_len, 0))) 
+	if (!(_l = insert_new_lump_after(_l, prefix, prefix_len, 0)))
 		goto lump_err;
 	prefix = 0;
-	_l = insert_subst_lump_after(_l, _inbound?SUBST_RCV_ALL:SUBST_SND_ALL, 0);
+	rr_lump_type = (_inbound)?
+					(rr_sockname_mode?SUBST_RCV_ALL_EX:SUBST_RCV_ALL)
+					:(rr_sockname_mode?SUBST_SND_ALL_EX:SUBST_SND_ALL);
+	_l = insert_subst_lump_after(_l, rr_lump_type, 0);
 	if (_l ==0 )
 		goto lump_err;
 	if (enable_double_rr) {
@@ -307,7 +313,7 @@ static inline int build_rr(struct lump* _l, struct lump* _l2, str* user,
 		goto lump_err;
 	term = 0;
 	return 0;
-	
+
 lump_err:
 	LM_ERR("failed to insert lumps\n");
 	if (prefix) pkg_free(prefix);
@@ -378,11 +384,11 @@ int record_route(struct sip_msg* _m, str *params)
 	str user = {NULL, 0};
 	str* tag;
 	int use_ob = rr_obb.use_outbound ? rr_obb.use_outbound(_m) : 0;
-	int sips;
+	int sips = 0;
 	int ret = 0;
-	
+
 	user.len = 0;
-	
+
 	if (add_username) {
 		/* check if there is a custom user set */
 		if (get_custom_user(_m, &user) < 0) {
@@ -392,7 +398,7 @@ int record_route(struct sip_msg* _m, str *params)
 			}
 		}
 	} else if (use_ob == 1) {
-		if (rr_obb.encode_flow_token(&user, _m->rcv) != 0) {
+		if (rr_obb.encode_flow_token(&user, &_m->rcv) != 0) {
 			LM_ERR("encoding outbound flow-token\n");
 			return -1;
 		}
@@ -428,7 +434,9 @@ int record_route(struct sip_msg* _m, str *params)
 		rr_param_buf.len = 0;
 	}
 
-	sips = rr_is_sips(_m);
+	if(rr_ignore_sips==0) {
+		sips = rr_is_sips(_m);
+	}
 
 	if (enable_double_rr) {
 		l = anchor_lump(_m, _m->headers->name.s - _m->buf,0,HDR_RECORDROUTE_T);
@@ -453,7 +461,7 @@ int record_route(struct sip_msg* _m, str *params)
 			goto error;
 		}
 	}
-	
+
 	l = anchor_lump(_m, _m->headers->name.s - _m->buf, 0, HDR_RECORDROUTE_T);
 	l2 = anchor_lump(_m, _m->headers->name.s - _m->buf, 0, 0);
 	if (!l || !l2) {
@@ -461,13 +469,15 @@ int record_route(struct sip_msg* _m, str *params)
 		ret = -3;
 		goto error;
 	}
-	
+
 	if (build_rr(l, l2, &user, tag, params, INBOUND, sips) < 0) {
 		LM_ERR("failed to insert inbound Record-Route\n");
 		ret = -4;
 		goto error;
 	}
 
+	LM_DBG("inserted record route (r2: %d - plen: %d)\n", enable_double_rr,
+			(params)?params->len:0);
 	/* reset the rr_param buffer */
 	rr_param_buf.len = 0;
 	ret = 0;
@@ -499,10 +509,12 @@ int record_route_preset(struct sip_msg* _m, str* _data)
 	int use_ob = rr_obb.use_outbound ? rr_obb.use_outbound(_m) : 0;
 	char *rr_prefix;
 	int rr_prefix_len;
-	int sips;
+	int sips = 0;
 	int ret = 0;
 
-	sips = rr_is_sips(_m);
+	if(rr_ignore_sips==0) {
+		sips = rr_is_sips(_m);
+	}
 	if(sips==0) {
 		rr_prefix = RR_PREFIX_SIP;
 		rr_prefix_len = RR_PREFIX_SIP_LEN;
@@ -517,7 +529,7 @@ int record_route_preset(struct sip_msg* _m, str* _data)
 			return -1;
 		}
 	} else if (use_ob == 1) {
-		if (rr_obb.encode_flow_token(&user, _m->rcv) != 0) {
+		if (rr_obb.encode_flow_token(&user, &_m->rcv) != 0) {
 			LM_ERR("encoding outbound flow-token\n");
 			return -1;
 		}
@@ -529,14 +541,19 @@ int record_route_preset(struct sip_msg* _m, str* _data)
 	}
 
 	if (append_fromtag) {
-		if (parse_from_header(_m) < 0) {
+		if (parse_from_header(_m) < 0 || get_from(_m)==NULL) {
 			LM_ERR("From parsing failed\n");
 			ret = -2;
 			goto error;
 		}
-		from = (struct to_body*)_m->from->parsed;
+		from = get_from(_m);
 	}
-	
+
+	if (rr_param_buf.len && rr_param_msg!=_m->id) {
+		/* rr_params were set for a different message -> reset buffer */
+		rr_param_buf.len = 0;
+	}
+
 	l = anchor_lump(_m, _m->headers->name.s - _m->buf, 0, HDR_RECORDROUTE_T);
 	if (!l) {
 		LM_ERR("failed to create lump anchor\n");
@@ -552,7 +569,7 @@ int record_route_preset(struct sip_msg* _m, str* _data)
 	if (append_fromtag && from->tag_value.len) {
 		hdr_len += RR_FROMTAG_LEN + from->tag_value.len;
 	}
-	
+
 	if (enable_full_lr) {
 		hdr_len += RR_LR_FULL_LEN;
 	} else {
@@ -581,7 +598,7 @@ int record_route_preset(struct sip_msg* _m, str* _data)
 
 	memcpy(p, _data->s, _data->len);
 	p += _data->len;
-	
+
 	if (append_fromtag && from->tag_value.len) {
 		memcpy(p, RR_FROMTAG, RR_FROMTAG_LEN);
 		p += RR_FROMTAG_LEN;
@@ -605,6 +622,7 @@ int record_route_preset(struct sip_msg* _m, str* _data)
 		ret = -5;
 		goto error;
 	}
+	LM_DBG("inserted preset record route\n");
 	ret = 1;
 error:
 	if ((use_ob == 1) || (use_ob == 2))
@@ -652,7 +670,7 @@ static inline int build_advertised_rr(struct lump* _l, struct lump* _l2, str *_d
 	if (tag && tag->len) {
 		suffix_len += RR_FROMTAG_LEN + tag->len;
 	}
-	
+
 	if (enable_full_lr) {
 		suffix_len += RR_LR_FULL_LEN;
 	} else {
@@ -686,7 +704,7 @@ static inline int build_advertised_rr(struct lump* _l, struct lump* _l2, str *_d
 	}
 
 	memcpy(p, _data->s, _data->len);
-	
+
 	p = suffix;
 	if (tag && tag->len) {
 		memcpy(p, RR_FROMTAG, RR_FROMTAG_LEN);
@@ -756,9 +774,9 @@ int record_route_advertised_address(struct sip_msg* _m, str* _data)
 	struct lump* l;
 	struct lump* l2;
 	int use_ob = rr_obb.use_outbound ? rr_obb.use_outbound(_m) : 0;
-	int sips;
+	int sips = 0;
 	int ret = 0;
-	
+
 	user.len = 0;
 	user.s = 0;
 
@@ -768,7 +786,7 @@ int record_route_advertised_address(struct sip_msg* _m, str* _data)
 			return -1;
 		}
 	} else if (use_ob == 1) {
-		if (rr_obb.encode_flow_token(&user, _m->rcv) != 0) {
+		if (rr_obb.encode_flow_token(&user, &_m->rcv) != 0) {
 			LM_ERR("encoding outbound flow-token\n");
 			return -1;
 		}
@@ -799,7 +817,14 @@ int record_route_advertised_address(struct sip_msg* _m, str* _data)
 		tag = 0;
 	}
 
-	sips = rr_is_sips(_m);
+	if (rr_param_buf.len && rr_param_msg!=_m->id) {
+		/* rr_params were set for a different message -> reset buffer */
+		rr_param_buf.len = 0;
+	}
+
+	if(rr_ignore_sips==0) {
+		sips = rr_is_sips(_m);
+	}
 
 	if (enable_double_rr) {
 		l = anchor_lump(_m, _m->headers->name.s - _m->buf,0,HDR_RECORDROUTE_T);
@@ -825,7 +850,7 @@ int record_route_advertised_address(struct sip_msg* _m, str* _data)
 			goto error;
 		}
 	}
-	
+
 	l = anchor_lump(_m, _m->headers->name.s - _m->buf, 0, HDR_RECORDROUTE_T);
 	l2 = anchor_lump(_m, _m->headers->name.s - _m->buf, 0, 0);
 	if (!l || !l2) {
@@ -833,12 +858,13 @@ int record_route_advertised_address(struct sip_msg* _m, str* _data)
 		ret = -6;
 		goto error;
 	}
-	
+
 	if (build_advertised_rr(l, l2, _data, &user, tag, INBOUND, sips) < 0) {
 		LM_ERR("failed to insert outbound Record-Route\n");
 		ret = -7;
 		goto error;
 	}
+	LM_DBG("inserted advertised address record route\n");
 	ret = 1;
 error:
 	if ((use_ob == 1) || (use_ob == 2))
@@ -862,7 +888,7 @@ static struct lump *get_rr_param_lump( struct lump** root)
 	for( crt=*root ; crt && !last ; crt=crt->next,(*root)=crt ) {
 		/* check on before list */
 		for( r=crt->before ; r ; r=r->before ) {
-			/* we are looking for the lump that adds the 
+			/* we are looking for the lump that adds the
 			 * suffix of the RR header */
 			if ( r->type==HDR_RECORDROUTE_T && r->op==LUMP_ADD)
 				last = r;
